from _balder.testresult import ResultState
from _balder.balder_session import BalderSession

from tests.test_utilities.base_0_envtester_class import Base0EnvtesterClass


class Test0ValidateVDevFeatReset(Base0EnvtesterClass):
    """
    This testcase executes the basic example and checks if the tree ends with the result SUCCESS.

    It has a vDevice-Device mapping on its scenario class and provides a vDevice definition only in the scenario
    feature (definition will not be repeated in setup feature). It has two scenarios and two applicable setups, while
    assigning different features to the internal VDevice feature. The test makes sure that VDevices can be overwritten
    by further setups.
    """

    @property
    def expected_data(self) -> tuple:
        return (
            # FIXTURE-CONSTRUCTION: balderglob_fixture_session
            {"file": "balderglob.py", "meth": "balderglob_fixture_session", "part": "construction"},
            [
                (
                    # FIXTURE-CONSTRUCTION: SetupBWorking.fixture_session
                    {"cls": "SetupBWorking", "meth": "fixture_session", "part": "construction"},
                    {"cls": "SetupFeatureI", "meth": "do_something", "category": "feature"},
                    {"cls": "SetupFeatureII", "meth": "do_something", "category": "feature"},
                ),
                (
                    # FIXTURE-CONSTRUCTION: SetupB2Working.fixture_session
                    {"cls": "SetupB2Working", "meth": "fixture_session", "part": "construction"},
                    {"cls": "SetupFeatureI2", "meth": "do_something", "category": "feature"},
                    {"cls": "SetupFeatureII2", "meth": "do_something", "category": "feature"},
                )
            ],
            [
                (
                    # FIXTURE-CONSTRUCTION: Scenario1.fixture_session
                    {"cls": "Scenario1", "meth": "fixture_session", "part": "construction"},
                    {"cls": "FeatureI", "meth": "do_something", "category": "feature"},
                    {"cls": "FeatureII", "meth": "do_something", "category": "feature"},
                ),
                (
                    # FIXTURE-CONSTRUCTION: Scenario2.fixture_session
                    {"cls": "Scenario2", "meth": "fixture_session", "part": "construction"},
                    {"cls": "FeatureI", "meth": "do_something", "category": "feature"},
                    {"cls": "FeatureII", "meth": "do_something", "category": "feature"},
                )
            ],
            [
                (
                    # FIXTURE-CONSTRUCTION: balderglob_fixture_setup
                    {"file": "balderglob.py", "meth": "balderglob_fixture_setup", "part": "construction"},
                    # FIXTURE-CONSTRUCTION: SetupBWorking.fixture_setup
                    {"cls": "SetupBWorking", "meth": "fixture_setup", "part": "construction"},
                    {"cls": "SetupFeatureI", "meth": "do_something", "category": "feature"},
                    {"cls": "SetupFeatureII", "meth": "do_something", "category": "feature"},
                    [
                        (
                            # FIXTURE-CONSTRUCTION: Scenario1.fixture_setup
                            {"cls": "Scenario1", "meth": "fixture_setup", "part": "construction"},
                            {"cls": "FeatureI", "meth": "do_something", "category": "feature"},
                            {"cls": "FeatureII", "meth": "do_something", "category": "feature"},
                        ),
                        (
                            # FIXTURE-CONSTRUCTION: Scenario2.fixture_setup
                            {"cls": "Scenario2", "meth": "fixture_setup", "part": "construction"},
                            {"cls": "FeatureI", "meth": "do_something", "category": "feature"},
                            {"cls": "FeatureII", "meth": "do_something", "category": "feature"},
                        )
                    ],
                    [
                        (
                            # FIXTURE-CONSTRUCTION: balderglob_fixture_scenario
                            {"file": "balderglob.py", "meth": "balderglob_fixture_scenario", "part": "construction"},
                            # FIXTURE-CONSTRUCTION: SetupBWorking.fixture_scenario
                            {"cls": "SetupBWorking", "meth": "fixture_scenario", "part": "construction"},
                            {"cls": "SetupFeatureI", "meth": "do_something", "category": "feature"},
                            {"cls": "SetupFeatureII", "meth": "do_something", "category": "feature"},
                            # FIXTURE-CONSTRUCTION: Scenario1.fixture_scenario
                            {"cls": "Scenario1", "meth": "fixture_scenario", "part": "construction"},
                            {"cls": "FeatureI", "meth": "do_something", "category": "feature"},
                            {"cls": "FeatureII", "meth": "do_something", "category": "feature"},
                            # FIXTURE-CONSTRUCTION: balderglob_fixture_variation
                            {"file": "balderglob.py", "meth": "balderglob_fixture_variation", "part": "construction"},
                            # FIXTURE-CONSTRUCTION: SetupBWorking.fixture_variation
                            {"cls": "SetupBWorking", "meth": "fixture_variation", "part": "construction"},
                            {"cls": "SetupFeatureI", "meth": "do_something", "category": "feature"},
                            {"cls": "SetupFeatureII", "meth": "do_something", "category": "feature"},
                            # FIXTURE-CONSTRUCTION: Scenario1.fixture_variation
                            {"cls": "Scenario1", "meth": "fixture_variation", "part": "construction"},
                            {"cls": "SetupFeatureI", "meth": "do_something", "category": "feature"},
                            {"cls": "SetupFeatureII", "meth": "do_something", "category": "feature"},
                            [
                                (
                                    # FIXTURE-CONSTRUCTION: balderglob_fixture_testcase
                                    {"file": "balderglob.py", "meth": "balderglob_fixture_testcase",
                                     "part": "construction"},
                                    # FIXTURE-CONSTRUCTION: SetupBWorking.fixture_testcase
                                    {"cls": "SetupBWorking", "meth": "fixture_testcase", "part": "construction"},
                                    {"cls": "SetupFeatureI", "meth": "do_something", "category": "feature"},
                                    {"cls": "SetupFeatureII", "meth": "do_something", "category": "feature"},
                                    # FIXTURE-CONSTRUCTION: Scenario1.fixture_testcase
                                    {"cls": "Scenario1", "meth": "fixture_testcase", "part": "construction"},
                                    {"cls": "SetupFeatureI", "meth": "do_something", "category": "feature"},
                                    {"cls": "SetupFeatureII", "meth": "do_something", "category": "feature"},
                                    # TESTCASE: Scenario1.test_scenario1_1
                                    {"cls": "Scenario1", "meth": "test_scenario1_1"},
                                    {"cls": "SetupFeatureI", "meth": "do_something", "category": "feature"},
                                    {"cls": "SetupFeatureII", "meth": "do_something", "category": "feature"},
                                    {"cls": "FeatureII", "meth": "do_something_special_with_the_vdevice",
                                     "category": "feature"},
                                    {"cls": "FeatureI", "meth": "do_something_special", "category": "feature"},
                                    # FIXTURE-TEARDOWN: Scenario1.fixture_testcase
                                    {"cls": "Scenario1", "meth": "fixture_testcase", "part": "teardown"},
                                    {"cls": "SetupFeatureI", "meth": "do_something", "category": "feature"},
                                    {"cls": "SetupFeatureII", "meth": "do_something", "category": "feature"},
                                    # FIXTURE-TEARDOWN: SetupBWorking.fixture_testcase
                                    {"cls": "SetupBWorking", "meth": "fixture_testcase", "part": "teardown"},
                                    {"cls": "SetupFeatureI", "meth": "do_something", "category": "feature"},
                                    {"cls": "SetupFeatureII", "meth": "do_something", "category": "feature"},
                                    # FIXTURE-TEARDOWN: balderglob_fixture_testcase
                                    {"file": "balderglob.py", "meth": "balderglob_fixture_testcase",
                                     "part": "teardown"},
                                ),
                                (
                                    # FIXTURE-CONSTRUCTION: balderglob_fixture_testcase
                                    {"file": "balderglob.py", "meth": "balderglob_fixture_testcase",
                                     "part": "construction"},
                                    # FIXTURE-CONSTRUCTION: SetupBWorking.fixture_testcase
                                    {"cls": "SetupBWorking", "meth": "fixture_testcase", "part": "construction"},
                                    {"cls": "SetupFeatureI", "meth": "do_something", "category": "feature"},
                                    {"cls": "SetupFeatureII", "meth": "do_something", "category": "feature"},
                                    # FIXTURE-CONSTRUCTION: Scenario1.fixture_testcase
                                    {"cls": "Scenario1", "meth": "fixture_testcase", "part": "construction"},
                                    {"cls": "SetupFeatureI", "meth": "do_something", "category": "feature"},
                                    {"cls": "SetupFeatureII", "meth": "do_something", "category": "feature"},
                                    # TESTCASE: Scenario1.test_scenario1_2
                                    {"cls": "Scenario1", "meth": "test_scenario1_2"},
                                    {"cls": "SetupFeatureI", "meth": "do_something", "category": "feature"},
                                    {"cls": "SetupFeatureII", "meth": "do_something", "category": "feature"},
                                    {"cls": "FeatureII", "meth": "do_something_special_with_the_vdevice",
                                     "category": "feature"},
                                    {"cls": "FeatureI", "meth": "do_something_special", "category": "feature"},
                                    # FIXTURE-TEARDOWN: Scenario1.fixture_testcase
                                    {"cls": "Scenario1", "meth": "fixture_testcase", "part": "teardown"},
                                    {"cls": "SetupFeatureI", "meth": "do_something", "category": "feature"},
                                    {"cls": "SetupFeatureII", "meth": "do_something", "category": "feature"},
                                    # FIXTURE-TEARDOWN: SetupBWorking.fixture_testcase
                                    {"cls": "SetupBWorking", "meth": "fixture_testcase", "part": "teardown"},
                                    {"cls": "SetupFeatureI", "meth": "do_something", "category": "feature"},
                                    {"cls": "SetupFeatureII", "meth": "do_something", "category": "feature"},
                                    # FIXTURE-TEARDOWN: balderglob_fixture_testcase
                                    {"file": "balderglob.py", "meth": "balderglob_fixture_testcase",
                                     "part": "teardown"},
                                )
                            ],
                            # FIXTURE-TEARDOWN: Scenario1.fixture_variation
                            {"cls": "Scenario1", "meth": "fixture_variation", "part": "teardown"},
                            {"cls": "SetupFeatureI", "meth": "do_something", "category": "feature"},
                            {"cls": "SetupFeatureII", "meth": "do_something", "category": "feature"},
                            # FIXTURE-TEARDOWN: SetupBWorking.fixture_variation
                            {"cls": "SetupBWorking", "meth": "fixture_variation", "part": "teardown"},
                            {"cls": "SetupFeatureI", "meth": "do_something", "category": "feature"},
                            {"cls": "SetupFeatureII", "meth": "do_something", "category": "feature"},
                            # FIXTURE-TEARDOWN: balderglob_fixture_variation
                            {"file": "balderglob.py", "meth": "balderglob_fixture_variation", "part": "teardown"},
                            # FIXTURE-TEARDOWN: Scenario1.fixture_scenario
                            {"cls": "Scenario1", "meth": "fixture_scenario", "part": "teardown"},
                            {"cls": "FeatureI", "meth": "do_something", "category": "feature"},
                            {"cls": "FeatureII", "meth": "do_something", "category": "feature"},
                            # FIXTURE-TEARDOWN: SetupBWorking.fixture_scenario
                            {"cls": "SetupBWorking", "meth": "fixture_scenario", "part": "teardown"},
                            {"cls": "SetupFeatureI", "meth": "do_something", "category": "feature"},
                            {"cls": "SetupFeatureII", "meth": "do_something", "category": "feature"},
                            # FIXTURE-TEARDOWN: balderglob_fixture_scenario
                            {"file": "balderglob.py", "meth": "balderglob_fixture_scenario", "part": "teardown"},
                        ),
                        (
                            # FIXTURE-CONSTRUCTION: balderglob_fixture_scenario
                            {"file": "balderglob.py", "meth": "balderglob_fixture_scenario", "part": "construction"},
                            # FIXTURE-CONSTRUCTION: SetupBWorking.fixture_scenario
                            {"cls": "SetupBWorking", "meth": "fixture_scenario", "part": "construction"},
                            {"cls": "SetupFeatureI", "meth": "do_something", "category": "feature"},
                            {"cls": "SetupFeatureII", "meth": "do_something", "category": "feature"},
                            # FIXTURE-CONSTRUCTION: Scenario2.fixture_scenario
                            {"cls": "Scenario2", "meth": "fixture_scenario", "part": "construction"},
                            {"cls": "FeatureI", "meth": "do_something", "category": "feature"},
                            {"cls": "FeatureII", "meth": "do_something", "category": "feature"},
                            # FIXTURE-CONSTRUCTION: balderglob_fixture_variation
                            {"file": "balderglob.py", "meth": "balderglob_fixture_variation", "part": "construction"},
                            # FIXTURE-CONSTRUCTION: SetupBWorking.fixture_variation
                            {"cls": "SetupBWorking", "meth": "fixture_variation", "part": "construction"},
                            {"cls": "SetupFeatureI", "meth": "do_something", "category": "feature"},
                            {"cls": "SetupFeatureII", "meth": "do_something", "category": "feature"},
                            # FIXTURE-CONSTRUCTION: Scenario2.fixture_variation
                            {"cls": "Scenario2", "meth": "fixture_variation", "part": "construction"},
                            {"cls": "SetupFeatureI", "meth": "do_something", "category": "feature"},
                            {"cls": "SetupFeatureII", "meth": "do_something", "category": "feature"},
                            [
                                (
                                    # FIXTURE-CONSTRUCTION: balderglob_fixture_testcase
                                    {"file": "balderglob.py", "meth": "balderglob_fixture_testcase",
                                     "part": "construction"},
                                    # FIXTURE-CONSTRUCTION: SetupBWorking.fixture_testcase
                                    {"cls": "SetupBWorking", "meth": "fixture_testcase", "part": "construction"},
                                    {"cls": "SetupFeatureI", "meth": "do_something", "category": "feature"},
                                    {"cls": "SetupFeatureII", "meth": "do_something", "category": "feature"},
                                    # FIXTURE-CONSTRUCTION: Scenario2.fixture_testcase
                                    {"cls": "Scenario2", "meth": "fixture_testcase", "part": "construction"},
                                    {"cls": "SetupFeatureI", "meth": "do_something", "category": "feature"},
                                    {"cls": "SetupFeatureII", "meth": "do_something", "category": "feature"},
                                    # TESTCASE: Scenario2.test_scenario2_1
                                    {"cls": "Scenario2", "meth": "test_scenario2_1"},
                                    {"cls": "SetupFeatureI", "meth": "do_something", "category": "feature"},
                                    {"cls": "SetupFeatureII", "meth": "do_something", "category": "feature"},
                                    {"cls": "FeatureII", "meth": "do_something_special_with_the_vdevice",
                                     "category": "feature"},
                                    {"cls": "FeatureI", "meth": "do_something_special", "category": "feature"},
                                    # FIXTURE-TEARDOWN: Scenario2.fixture_testcase
                                    {"cls": "Scenario2", "meth": "fixture_testcase", "part": "teardown"},
                                    {"cls": "SetupFeatureI", "meth": "do_something", "category": "feature"},
                                    {"cls": "SetupFeatureII", "meth": "do_something", "category": "feature"},
                                    # FIXTURE-TEARDOWN: SetupBWorking.fixture_testcase
                                    {"cls": "SetupBWorking", "meth": "fixture_testcase", "part": "teardown"},
                                    {"cls": "SetupFeatureI", "meth": "do_something", "category": "feature"},
                                    {"cls": "SetupFeatureII", "meth": "do_something", "category": "feature"},
                                    # FIXTURE-TEARDOWN: balderglob_fixture_testcase
                                    {"file": "balderglob.py", "meth": "balderglob_fixture_testcase",
                                     "part": "teardown"},
                                ),
                                (
                                    # FIXTURE-CONSTRUCTION: balderglob_fixture_testcase
                                    {"file": "balderglob.py", "meth": "balderglob_fixture_testcase",
                                     "part": "construction"},
                                    # FIXTURE-CONSTRUCTION: SetupBWorking.fixture_testcase
                                    {"cls": "SetupBWorking", "meth": "fixture_testcase", "part": "construction"},
                                    {"cls": "SetupFeatureI", "meth": "do_something", "category": "feature"},
                                    {"cls": "SetupFeatureII", "meth": "do_something", "category": "feature"},
                                    # FIXTURE-CONSTRUCTION: Scenario2.fixture_testcase
                                    {"cls": "Scenario2", "meth": "fixture_testcase", "part": "construction"},
                                    {"cls": "SetupFeatureI", "meth": "do_something", "category": "feature"},
                                    {"cls": "SetupFeatureII", "meth": "do_something", "category": "feature"},
                                    # TESTCASE: Scenario2.test_scenario2_2
                                    {"cls": "Scenario2", "meth": "test_scenario2_2"},
                                    {"cls": "SetupFeatureI", "meth": "do_something", "category": "feature"},
                                    {"cls": "SetupFeatureII", "meth": "do_something", "category": "feature"},
                                    {"cls": "FeatureII", "meth": "do_something_special_with_the_vdevice",
                                     "category": "feature"},
                                    {"cls": "FeatureI", "meth": "do_something_special", "category": "feature"},
                                    # FIXTURE-TEARDOWN: Scenario2.fixture_testcase
                                    {"cls": "Scenario2", "meth": "fixture_testcase", "part": "teardown"},
                                    {"cls": "SetupFeatureI", "meth": "do_something", "category": "feature"},
                                    {"cls": "SetupFeatureII", "meth": "do_something", "category": "feature"},
                                    # FIXTURE-TEARDOWN: SetupBWorking.fixture_testcase
                                    {"cls": "SetupBWorking", "meth": "fixture_testcase", "part": "teardown"},
                                    {"cls": "SetupFeatureI", "meth": "do_something", "category": "feature"},
                                    {"cls": "SetupFeatureII", "meth": "do_something", "category": "feature"},
                                    # FIXTURE-TEARDOWN: balderglob_fixture_testcase
                                    {"file": "balderglob.py", "meth": "balderglob_fixture_testcase",
                                     "part": "teardown"},
                                )
                            ],
                            # FIXTURE-TEARDOWN: Scenario2.fixture_variation
                            {"cls": "Scenario2", "meth": "fixture_variation", "part": "teardown"},
                            {"cls": "SetupFeatureI", "meth": "do_something", "category": "feature"},
                            {"cls": "SetupFeatureII", "meth": "do_something", "category": "feature"},
                            # FIXTURE-TEARDOWN: SetupBWorking.fixture_variation
                            {"cls": "SetupBWorking", "meth": "fixture_variation", "part": "teardown"},
                            {"cls": "SetupFeatureI", "meth": "do_something", "category": "feature"},
                            {"cls": "SetupFeatureII", "meth": "do_something", "category": "feature"},
                            # FIXTURE-TEARDOWN: balderglob_fixture_variation
                            {"file": "balderglob.py", "meth": "balderglob_fixture_variation", "part": "teardown"},
                            # FIXTURE-TEARDOWN: Scenario2.fixture_scenario
                            {"cls": "Scenario2", "meth": "fixture_scenario", "part": "teardown"},
                            {"cls": "FeatureI", "meth": "do_something", "category": "feature"},
                            {"cls": "FeatureII", "meth": "do_something", "category": "feature"},
                            # FIXTURE-TEARDOWN: SetupBWorking.fixture_scenario
                            {"cls": "SetupBWorking", "meth": "fixture_scenario", "part": "teardown"},
                            {"cls": "SetupFeatureI", "meth": "do_something", "category": "feature"},
                            {"cls": "SetupFeatureII", "meth": "do_something", "category": "feature"},
                            # FIXTURE-TEARDOWN: balderglob_fixture_scenario
                            {"file": "balderglob.py", "meth": "balderglob_fixture_scenario", "part": "teardown"},
                        )
                    ],
                    [
                        (
                            # FIXTURE-TEARDOWN: Scenario1.fixture_setup
                            {"cls": "Scenario1", "meth": "fixture_setup", "part": "teardown"},
                            {"cls": "FeatureI", "meth": "do_something", "category": "feature"},
                            {"cls": "FeatureII", "meth": "do_something", "category": "feature"},
                        ),
                        (
                            # FIXTURE-TEARDOWN: Scenario2.fixture_setup
                            {"cls": "Scenario2", "meth": "fixture_setup", "part": "teardown"},
                            {"cls": "FeatureI", "meth": "do_something", "category": "feature"},
                            {"cls": "FeatureII", "meth": "do_something", "category": "feature"},
                        )
                    ],
                    # FIXTURE-TEARDOWN: SetupBWorking.fixture_setup
                    {"cls": "SetupBWorking", "meth": "fixture_setup", "part": "teardown"},
                    {"cls": "SetupFeatureI", "meth": "do_something", "category": "feature"},
                    {"cls": "SetupFeatureII", "meth": "do_something", "category": "feature"},
                    # FIXTURE-TEARDOWN: balderglob_fixture_setup
                    {"file": "balderglob.py", "meth": "balderglob_fixture_setup", "part": "teardown"},
                ),
                (
                    # FIXTURE-CONSTRUCTION: balderglob_fixture_setup
                    {"file": "balderglob.py", "meth": "balderglob_fixture_setup", "part": "construction"},
                    # FIXTURE-CONSTRUCTION: SetupB2Working.fixture_setup
                    {"cls": "SetupB2Working", "meth": "fixture_setup", "part": "construction"},
                    {"cls": "SetupFeatureI2", "meth": "do_something", "category": "feature"},
                    {"cls": "SetupFeatureII2", "meth": "do_something", "category": "feature"},
                    [
                        (
                            # FIXTURE-CONSTRUCTION: Scenario1.fixture_setup
                            {"cls": "Scenario1", "meth": "fixture_setup", "part": "construction"},
                            {"cls": "FeatureI", "meth": "do_something", "category": "feature"},
                            {"cls": "FeatureII", "meth": "do_something", "category": "feature"},
                        ),
                        (
                            # FIXTURE-CONSTRUCTION: Scenario2.fixture_setup
                            {"cls": "Scenario2", "meth": "fixture_setup", "part": "construction"},
                            {"cls": "FeatureI", "meth": "do_something", "category": "feature"},
                            {"cls": "FeatureII", "meth": "do_something", "category": "feature"},
                        )
                    ],
                    [
                        (
                            # FIXTURE-CONSTRUCTION: balderglob_fixture_scenario
                            {"file": "balderglob.py", "meth": "balderglob_fixture_scenario", "part": "construction"},
                            # FIXTURE-CONSTRUCTION: SetupB2Working.fixture_scenario
                            {"cls": "SetupB2Working", "meth": "fixture_scenario", "part": "construction"},
                            {"cls": "SetupFeatureI2", "meth": "do_something", "category": "feature"},
                            {"cls": "SetupFeatureII2", "meth": "do_something", "category": "feature"},
                            # FIXTURE-CONSTRUCTION: Scenario1.fixture_scenario
                            {"cls": "Scenario1", "meth": "fixture_scenario", "part": "construction"},
                            {"cls": "FeatureI", "meth": "do_something", "category": "feature"},
                            {"cls": "FeatureII", "meth": "do_something", "category": "feature"},
                            # FIXTURE-CONSTRUCTION: balderglob_fixture_variation
                            {"file": "balderglob.py", "meth": "balderglob_fixture_variation", "part": "construction"},
                            # FIXTURE-CONSTRUCTION: SetupB2Working.fixture_variation
                            {"cls": "SetupB2Working", "meth": "fixture_variation", "part": "construction"},
                            {"cls": "SetupFeatureI2", "meth": "do_something", "category": "feature"},
                            {"cls": "SetupFeatureII2", "meth": "do_something", "category": "feature"},
                            # FIXTURE-CONSTRUCTION: Scenario1.fixture_variation
                            {"cls": "Scenario1", "meth": "fixture_variation", "part": "construction"},
                            {"cls": "SetupFeatureI2", "meth": "do_something", "category": "feature"},
                            {"cls": "SetupFeatureII2", "meth": "do_something", "category": "feature"},
                            [
                                (
                                    # FIXTURE-CONSTRUCTION: balderglob_fixture_testcase
                                    {"file": "balderglob.py", "meth": "balderglob_fixture_testcase", "part": "construction"},
                                    # FIXTURE-CONSTRUCTION: SetupB2Working.fixture_testcase
                                    {"cls": "SetupB2Working", "meth": "fixture_testcase", "part": "construction"},
                                    {"cls": "SetupFeatureI2", "meth": "do_something", "category": "feature"},
                                    {"cls": "SetupFeatureII2", "meth": "do_something", "category": "feature"},
                                    # FIXTURE-CONSTRUCTION: Scenario1.fixture_testcase
                                    {"cls": "Scenario1", "meth": "fixture_testcase", "part": "construction"},
                                    {"cls": "SetupFeatureI2", "meth": "do_something", "category": "feature"},
                                    {"cls": "SetupFeatureII2", "meth": "do_something", "category": "feature"},
                                    # TESTCASE: Scenario1.test_scenario1_1
                                    {"cls": "Scenario1", "meth": "test_scenario1_1"},
                                    {"cls": "SetupFeatureI2", "meth": "do_something", "category": "feature"},
                                    {"cls": "SetupFeatureII2", "meth": "do_something", "category": "feature"},
                                    {"cls": "FeatureII", "meth": "do_something_special_with_the_vdevice", "category": "feature"},
                                    {"cls": "FeatureI", "meth": "do_something_special", "category": "feature"},
                                    # FIXTURE-TEARDOWN: Scenario1.fixture_testcase
                                    {"cls": "Scenario1", "meth": "fixture_testcase", "part": "teardown"},
                                    {"cls": "SetupFeatureI2", "meth": "do_something", "category": "feature"},
                                    {"cls": "SetupFeatureII2", "meth": "do_something", "category": "feature"},
                                    # FIXTURE-TEARDOWN: SetupB2Working.fixture_testcase
                                    {"cls": "SetupB2Working", "meth": "fixture_testcase", "part": "teardown"},
                                    {"cls": "SetupFeatureI2", "meth": "do_something", "category": "feature"},
                                    {"cls": "SetupFeatureII2", "meth": "do_something", "category": "feature"},
                                    # FIXTURE-TEARDOWN: balderglob_fixture_testcase
                                    {"file": "balderglob.py", "meth": "balderglob_fixture_testcase",
                                     "part": "teardown"},
                                ),
                                (
                                    # FIXTURE-CONSTRUCTION: balderglob_fixture_testcase
                                    {"file": "balderglob.py", "meth": "balderglob_fixture_testcase",
                                     "part": "construction"},
                                    # FIXTURE-CONSTRUCTION: SetupB2Working.fixture_testcase
                                    {"cls": "SetupB2Working", "meth": "fixture_testcase", "part": "construction"},
                                    {"cls": "SetupFeatureI2", "meth": "do_something", "category": "feature"},
                                    {"cls": "SetupFeatureII2", "meth": "do_something", "category": "feature"},
                                    # FIXTURE-CONSTRUCTION: Scenario1.fixture_testcase
                                    {"cls": "Scenario1", "meth": "fixture_testcase", "part": "construction"},
                                    {"cls": "SetupFeatureI2", "meth": "do_something", "category": "feature"},
                                    {"cls": "SetupFeatureII2", "meth": "do_something", "category": "feature"},
                                    # TESTCASE: Scenario1.test_scenario1_2
                                    {"cls": "Scenario1", "meth": "test_scenario1_2"},
                                    {"cls": "SetupFeatureI2", "meth": "do_something", "category": "feature"},
                                    {"cls": "SetupFeatureII2", "meth": "do_something", "category": "feature"},
                                    {"cls": "FeatureII", "meth": "do_something_special_with_the_vdevice",
                                     "category": "feature"},
                                    {"cls": "FeatureI", "meth": "do_something_special", "category": "feature"},
                                    # FIXTURE-TEARDOWN: Scenario1.fixture_testcase
                                    {"cls": "Scenario1", "meth": "fixture_testcase", "part": "teardown"},
                                    {"cls": "SetupFeatureI2", "meth": "do_something", "category": "feature"},
                                    {"cls": "SetupFeatureII2", "meth": "do_something", "category": "feature"},
                                    # FIXTURE-TEARDOWN: SetupB2Working.fixture_testcase
                                    {"cls": "SetupB2Working", "meth": "fixture_testcase", "part": "teardown"},
                                    {"cls": "SetupFeatureI2", "meth": "do_something", "category": "feature"},
                                    {"cls": "SetupFeatureII2", "meth": "do_something", "category": "feature"},
                                    # FIXTURE-TEARDOWN: balderglob_fixture_testcase
                                    {"file": "balderglob.py", "meth": "balderglob_fixture_testcase",
                                     "part": "teardown"},
                                )
                            ],
                            # FIXTURE-TEARDOWN: Scenario1.fixture_variation
                            {"cls": "Scenario1", "meth": "fixture_variation", "part": "teardown"},
                            {"cls": "SetupFeatureI2", "meth": "do_something", "category": "feature"},
                            {"cls": "SetupFeatureII2", "meth": "do_something", "category": "feature"},
                            # FIXTURE-TEARDOWN: SetupB2Working.fixture_variation
                            {"cls": "SetupB2Working", "meth": "fixture_variation", "part": "teardown"},
                            {"cls": "SetupFeatureI2", "meth": "do_something", "category": "feature"},
                            {"cls": "SetupFeatureII2", "meth": "do_something", "category": "feature"},
                            # FIXTURE-TEARDOWN: balderglob_fixture_variation
                            {"file": "balderglob.py", "meth": "balderglob_fixture_variation", "part": "teardown"},
                            # FIXTURE-TEARDOWN: Scenario1.fixture_scenario
                            {"cls": "Scenario1", "meth": "fixture_scenario", "part": "teardown"},
                            {"cls": "FeatureI", "meth": "do_something", "category": "feature"},
                            {"cls": "FeatureII", "meth": "do_something", "category": "feature"},
                            # FIXTURE-TEARDOWN: SetupB2Working.fixture_scenario
                            {"cls": "SetupB2Working", "meth": "fixture_scenario", "part": "teardown"},
                            {"cls": "SetupFeatureI2", "meth": "do_something", "category": "feature"},
                            {"cls": "SetupFeatureII2", "meth": "do_something", "category": "feature"},
                            # FIXTURE-TEARDOWN: balderglob_fixture_scenario
                            {"file": "balderglob.py", "meth": "balderglob_fixture_scenario", "part": "teardown"},
                        ),
                        (
                            # FIXTURE-CONSTRUCTION: balderglob_fixture_scenario
                            {"file": "balderglob.py", "meth": "balderglob_fixture_scenario", "part": "construction"},
                            # FIXTURE-CONSTRUCTION: SetupB2Working.fixture_scenario
                            {"cls": "SetupB2Working", "meth": "fixture_scenario", "part": "construction"},
                            {"cls": "SetupFeatureI2", "meth": "do_something", "category": "feature"},
                            {"cls": "SetupFeatureII2", "meth": "do_something", "category": "feature"},
                            # FIXTURE-CONSTRUCTION: Scenario2.fixture_scenario
                            {"cls": "Scenario2", "meth": "fixture_scenario", "part": "construction"},
                            {"cls": "FeatureI", "meth": "do_something", "category": "feature"},
                            {"cls": "FeatureII", "meth": "do_something", "category": "feature"},
                            # FIXTURE-CONSTRUCTION: balderglob_fixture_variation
                            {"file": "balderglob.py", "meth": "balderglob_fixture_variation", "part": "construction"},
                            # FIXTURE-CONSTRUCTION: SetupB2Working.fixture_variation
                            {"cls": "SetupB2Working", "meth": "fixture_variation", "part": "construction"},
                            {"cls": "SetupFeatureI2", "meth": "do_something", "category": "feature"},
                            {"cls": "SetupFeatureII2", "meth": "do_something", "category": "feature"},
                            # FIXTURE-CONSTRUCTION: Scenario2.fixture_variation
                            {"cls": "Scenario2", "meth": "fixture_variation", "part": "construction"},
                            {"cls": "SetupFeatureI2", "meth": "do_something", "category": "feature"},
                            {"cls": "SetupFeatureII2", "meth": "do_something", "category": "feature"},
                            [
                                (
                                    # FIXTURE-CONSTRUCTION: balderglob_fixture_testcase
                                    {"file": "balderglob.py", "meth": "balderglob_fixture_testcase",
                                     "part": "construction"},
                                    # FIXTURE-CONSTRUCTION: SetupB2Working.fixture_testcase
                                    {"cls": "SetupB2Working", "meth": "fixture_testcase", "part": "construction"},
                                    {"cls": "SetupFeatureI2", "meth": "do_something", "category": "feature"},
                                    {"cls": "SetupFeatureII2", "meth": "do_something", "category": "feature"},
                                    # FIXTURE-CONSTRUCTION: Scenario2.fixture_testcase
                                    {"cls": "Scenario2", "meth": "fixture_testcase", "part": "construction"},
                                    {"cls": "SetupFeatureI2", "meth": "do_something", "category": "feature"},
                                    {"cls": "SetupFeatureII2", "meth": "do_something", "category": "feature"},
                                    # TESTCASE: Scenario2.test_scenario2_1
                                    {"cls": "Scenario2", "meth": "test_scenario2_1"},
                                    {"cls": "SetupFeatureI2", "meth": "do_something", "category": "feature"},
                                    {"cls": "SetupFeatureII2", "meth": "do_something", "category": "feature"},
                                    {"cls": "FeatureII", "meth": "do_something_special_with_the_vdevice", "category": "feature"},
                                    {"cls": "FeatureI", "meth": "do_something_special", "category": "feature"},
                                    # FIXTURE-TEARDOWN: Scenario2.fixture_testcase
                                    {"cls": "Scenario2", "meth": "fixture_testcase", "part": "teardown"},
                                    {"cls": "SetupFeatureI2", "meth": "do_something", "category": "feature"},
                                    {"cls": "SetupFeatureII2", "meth": "do_something", "category": "feature"},
                                    # FIXTURE-TEARDOWN: SetupB2Working.fixture_testcase
                                    {"cls": "SetupB2Working", "meth": "fixture_testcase", "part": "teardown"},
                                    {"cls": "SetupFeatureI2", "meth": "do_something", "category": "feature"},
                                    {"cls": "SetupFeatureII2", "meth": "do_something", "category": "feature"},
                                    # FIXTURE-TEARDOWN: balderglob_fixture_testcase
                                    {"file": "balderglob.py", "meth": "balderglob_fixture_testcase", "part": "teardown"},
                                ),
                                (
                                    # FIXTURE-CONSTRUCTION: balderglob_fixture_testcase
                                    {"file": "balderglob.py", "meth": "balderglob_fixture_testcase", "part": "construction"},
                                    # FIXTURE-CONSTRUCTION: SetupB2Working.fixture_testcase
                                    {"cls": "SetupB2Working", "meth": "fixture_testcase", "part": "construction"},
                                    {"cls": "SetupFeatureI2", "meth": "do_something", "category": "feature"},
                                    {"cls": "SetupFeatureII2", "meth": "do_something", "category": "feature"},
                                    # FIXTURE-CONSTRUCTION: Scenario2.fixture_testcase
                                    {"cls": "Scenario2", "meth": "fixture_testcase", "part": "construction"},
                                    {"cls": "SetupFeatureI2", "meth": "do_something", "category": "feature"},
                                    {"cls": "SetupFeatureII2", "meth": "do_something", "category": "feature"},
                                    # TESTCASE: Scenario2.test_scenario2_2
                                    {"cls": "Scenario2", "meth": "test_scenario2_2"},
                                    {"cls": "SetupFeatureI2", "meth": "do_something", "category": "feature"},
                                    {"cls": "SetupFeatureII2", "meth": "do_something", "category": "feature"},
                                    {"cls": "FeatureII", "meth": "do_something_special_with_the_vdevice", "category": "feature"},
                                    {"cls": "FeatureI", "meth": "do_something_special", "category": "feature"},
                                    # FIXTURE-TEARDOWN: Scenario2.fixture_testcase
                                    {"cls": "Scenario2", "meth": "fixture_testcase", "part": "teardown"},
                                    {"cls": "SetupFeatureI2", "meth": "do_something", "category": "feature"},
                                    {"cls": "SetupFeatureII2", "meth": "do_something", "category": "feature"},
                                    # FIXTURE-TEARDOWN: SetupB2Working.fixture_testcase
                                    {"cls": "SetupB2Working", "meth": "fixture_testcase", "part": "teardown"},
                                    {"cls": "SetupFeatureI2", "meth": "do_something", "category": "feature"},
                                    {"cls": "SetupFeatureII2", "meth": "do_something", "category": "feature"},
                                    # FIXTURE-TEARDOWN: balderglob_fixture_testcase
                                    {"file": "balderglob.py", "meth": "balderglob_fixture_testcase", "part": "teardown"},
                                )
                            ],
                            # FIXTURE-TEARDOWN: Scenario2.fixture_variation
                            {"cls": "Scenario2", "meth": "fixture_variation", "part": "teardown"},
                            {"cls": "SetupFeatureI2", "meth": "do_something", "category": "feature"},
                            {"cls": "SetupFeatureII2", "meth": "do_something", "category": "feature"},
                            # FIXTURE-TEARDOWN: SetupB2Working.fixture_variation
                            {"cls": "SetupB2Working", "meth": "fixture_variation", "part": "teardown"},
                            {"cls": "SetupFeatureI2", "meth": "do_something", "category": "feature"},
                            {"cls": "SetupFeatureII2", "meth": "do_something", "category": "feature"},
                            # FIXTURE-TEARDOWN: balderglob_fixture_variation
                            {"file": "balderglob.py", "meth": "balderglob_fixture_variation", "part": "teardown"},
                            # FIXTURE-TEARDOWN: Scenario2.fixture_scenario
                            {"cls": "Scenario2", "meth": "fixture_scenario", "part": "teardown"},
                            {"cls": "FeatureI", "meth": "do_something", "category": "feature"},
                            {"cls": "FeatureII", "meth": "do_something", "category": "feature"},
                            # FIXTURE-TEARDOWN: SetupB2Working.fixture_scenario
                            {"cls": "SetupB2Working", "meth": "fixture_scenario", "part": "teardown"},
                            {"cls": "SetupFeatureI2", "meth": "do_something", "category": "feature"},
                            {"cls": "SetupFeatureII2", "meth": "do_something", "category": "feature"},
                            # FIXTURE-TEARDOWN: balderglob_fixture_scenario
                            {"file": "balderglob.py", "meth": "balderglob_fixture_scenario", "part": "teardown"},
                        )
                    ],
                    [
                        (
                            # FIXTURE-TEARDOWN: Scenario1.fixture_setup
                            {"cls": "Scenario1", "meth": "fixture_setup", "part": "teardown"},
                            {"cls": "FeatureI", "meth": "do_something", "category": "feature"},
                            {"cls": "FeatureII", "meth": "do_something", "category": "feature"},
                        ),
                        (
                            # FIXTURE-TEARDOWN: Scenario2.fixture_setup
                            {"cls": "Scenario2", "meth": "fixture_setup", "part": "teardown"},
                            {"cls": "FeatureI", "meth": "do_something", "category": "feature"},
                            {"cls": "FeatureII", "meth": "do_something", "category": "feature"},
                        )
                    ],
                    # FIXTURE-TEARDOWN: SetupB2Working.fixture_setup
                    {"cls": "SetupB2Working", "meth": "fixture_setup", "part": "teardown"},
                    {"cls": "SetupFeatureI2", "meth": "do_something", "category": "feature"},
                    {"cls": "SetupFeatureII2", "meth": "do_something", "category": "feature"},
                    # FIXTURE-TEARDOWN: balderglob_fixture_setup
                    {"file": "balderglob.py", "meth": "balderglob_fixture_setup", "part": "teardown"},
                )
            ],
            [
                (
                    # FIXTURE-TEARDOWN: Scenario1.fixture_session
                    {"cls": "Scenario1", "meth": "fixture_session", "part": "teardown"},
                    {"cls": "FeatureI", "meth": "do_something", "category": "feature"},
                    {"cls": "FeatureII", "meth": "do_something", "category": "feature"},
                ),
                (
                    # FIXTURE-TEARDOWN: Scenario2.fixture_session
                    {"cls": "Scenario2", "meth": "fixture_session", "part": "teardown"},
                    {"cls": "FeatureI", "meth": "do_something", "category": "feature"},
                    {"cls": "FeatureII", "meth": "do_something", "category": "feature"},
                )
            ],
            [
                (
                    # FIXTURE-TEARDOWN: SetupBWorking.fixture_session
                    {"cls": "SetupBWorking", "meth": "fixture_session", "part": "teardown"},
                    {"cls": "SetupFeatureI", "meth": "do_something", "category": "feature"},
                    {"cls": "SetupFeatureII", "meth": "do_something", "category": "feature"},
                ),
                (
                    # FIXTURE-TEARDOWN: SetupB2Working.fixture_session
                    {"cls": "SetupB2Working", "meth": "fixture_session", "part": "teardown"},
                    {"cls": "SetupFeatureI2", "meth": "do_something", "category": "feature"},
                    {"cls": "SetupFeatureII2", "meth": "do_something", "category": "feature"},
                )
            ],
            # FIXTURE-TEARDOWN: balderglob_fixture_session
            {"file": "balderglob.py", "meth": "balderglob_fixture_session", "part": "teardown"},
        )

    @staticmethod
    def validate_finished_session(session: BalderSession):
        # check result states everywhere (have to be SUCCESS everywhere)
        assert session.executor_tree.executor_result == ResultState.SUCCESS, \
            "test session does not terminates with success"

        assert session.executor_tree.construct_result.result == ResultState.SUCCESS, \
            "global executor tree construct part does not set ResultState.SUCCESS"
        assert session.executor_tree.body_result.result == ResultState.SUCCESS, \
            "global executor tree body part does not set ResultState.SUCCESS"
        assert session.executor_tree.teardown_result.result == ResultState.SUCCESS, \
            "global executor tree teardown part does not set ResultState.SUCCESS"
        for cur_setup_executor in session.executor_tree.get_setup_executors():
            assert cur_setup_executor.executor_result == ResultState.SUCCESS, \
                "the setup executor does not have result SUCCESS"

            assert cur_setup_executor.construct_result.result == ResultState.SUCCESS
            assert cur_setup_executor.body_result.result == ResultState.SUCCESS
            assert cur_setup_executor.teardown_result.result == ResultState.SUCCESS

            for cur_scenario_executor in cur_setup_executor.get_scenario_executors():
                assert cur_scenario_executor.executor_result == ResultState.SUCCESS, \
                    "the scenario executor does not have result SUCCESS"

                assert cur_scenario_executor.construct_result.result == ResultState.SUCCESS
                assert cur_scenario_executor.body_result.result == ResultState.SUCCESS
                assert cur_scenario_executor.teardown_result.result == ResultState.SUCCESS

                for cur_variation_executor in cur_scenario_executor.get_variation_executors():
                    assert cur_variation_executor.executor_result == ResultState.SUCCESS, \
                        "the variation executor does not have result SUCCESS"

                    assert cur_variation_executor.construct_result.result == ResultState.SUCCESS
                    assert cur_variation_executor.body_result.result == ResultState.SUCCESS
                    assert cur_variation_executor.teardown_result.result == ResultState.SUCCESS

                    for cur_testcase_executor in cur_variation_executor.get_testcase_executors():
                        assert cur_testcase_executor.executor_result == ResultState.SUCCESS, \
                            "the testcase executor does not have result SUCCESS"

                        assert cur_testcase_executor.construct_result.result == ResultState.SUCCESS
                        assert cur_testcase_executor.body_result.result == ResultState.SUCCESS
                        assert cur_testcase_executor.teardown_result.result == ResultState.SUCCESS



from balder.exceptions import DeviceOverwritingError, BalderException
from _balder.balder_session import BalderSession

from tests.test_utilities.base_0_envtester_class import Base0EnvtesterClass


class Test0ScenarioInheritanceOverwriteOnlyOneDevice(Base0EnvtesterClass):
    """
    This testcase executes a reduced version of the basic envtester environment. It only implements the `ScenarioA` and
    its related `SetupA` and a child class of the related ScenarioA that has an additional test method.

    The child class DOES ONLY overwrite one device instead of two. This should be forbidden. The collector should throw
    an exception.

    .. note::
        The `ScenarioAParent` class has a child (the `ScenarioAChild` class). This forbids the execution of the
        `ScenarioAParent` class.
    """
    @property
    def expected_data(self) -> tuple:
        return ()

    @property
    def expected_exit_code(self) -> int:
        return 4

    @staticmethod
    def handle_balder_exception(exc: BalderException):
        assert isinstance(exc, DeviceOverwritingError), 'unexpected exception type'
        assert exc.args[0] == "found a device `ScenarioAParent.ScenarioDevice2` which is part of a parent class, " \
                              "but it is not implemented in child class `ScenarioAChild`"

    @staticmethod
    def validate_finished_session(session: BalderSession):
        assert session.executor_tree is None, "test session does not terminates before collector work was done"




from _balder.testresult import ResultState
from _balder.balder_session import BalderSession

from tests.test_utilities.base_0_envtester_class import Base0EnvtesterClass


class Test0ExactlyOneLessFeatureInSetup(Base0EnvtesterClass):
    """
    This testcase uses a reduced version of the basic envtester environment. It should have no matching between the
    ``ScenarioA`` and the ``SetupA`` because the ``SetupDevice1`` does not provide an implementation of ``FeatureIII``.
    """

    @property
    def expected_data(self) -> tuple:
        return ()

    @staticmethod
    def validate_finished_session(session: BalderSession):
        # check result states everywhere (have to be SUCCESS everywhere
        assert session.executor_tree.executor_result == ResultState.NOT_RUN, \
            "test session does not terminates with NOT_RUN"

        assert session.executor_tree.construct_result.result == ResultState.NOT_RUN, \
            "global executor tree construct part does not set ResultState.NOT_RUN"
        assert session.executor_tree.body_result.result == ResultState.NOT_RUN, \
            "global executor tree body part does not set ResultState.SUCCESS"
        assert session.executor_tree.teardown_result.result == ResultState.NOT_RUN, \
            "global executor tree teardown part does not set ResultState.NOT_RUN"



from _balder import __version__, __version_tuple__
from _balder.setup import Setup
from _balder.device import Device
from _balder.vdevice import VDevice
from _balder.feature import Feature
from _balder.scenario import Scenario
from _balder.connection import Connection
from _balder.balder_plugin import BalderPlugin
from _balder.balder_settings import BalderSettings
from _balder.unmapped_vdevice import UnmappedVDevice
from _balder.decorator_fixture import fixture
from _balder.decorator_connect import connect
from _balder.decorator_covered_by import covered_by
from _balder.decorator_for_vdevice import for_vdevice
from _balder.decorator_insert_into_tree import insert_into_tree
from _balder.decorator_parametrize import parametrize
from _balder.decorator_parametrize_by_feature import parametrize_by_feature


__all__ = [
    '__version__',

    '__version_tuple__',

    'connect',

    'fixture',

    'for_vdevice',

    'insert_into_tree',

    'covered_by',

    'parametrize',

    'parametrize_by_feature',

    'Setup',

    'Device',

    'VDevice',

    'Feature',

    'Scenario',

    'Connection',

    'BalderPlugin',

    'BalderSettings',

    'UnmappedVDevice'
]



from __future__ import annotations

import time
from typing import TYPE_CHECKING

import sys

import traceback
from abc import ABC, abstractmethod

from _balder.executor.basic_executor import BasicExecutor
from _balder.testresult import ResultState

if TYPE_CHECKING:
    from _balder.fixture_execution_level import FixtureExecutionLevel
    from _balder.fixture_manager import FixtureManager


class BasicExecutableExecutor(BasicExecutor, ABC):
    """
    The BasicExecutor class is an abstract class that represents the parent class of all executors. Together with other
    executor classes, an executor forms a tree structure in which individual tests, which later on are executed, are
    assigned to individual scenarios
    """
    fixture_execution_level: FixtureExecutionLevel = None

    def __init__(self):
        super().__init__()

        # holds the execution time of this branch (with branch fixtures)
        self.execution_time_sec = None

    # ---------------------------------- STATIC METHODS ----------------------------------------------------------------

    # ---------------------------------- CLASS METHODS ----------------------------------------------------------------

    # ---------------------------------- PROPERTIES --------------------------------------------------------------------

    @property
    @abstractmethod
    def fixture_manager(self) -> FixtureManager:
        """returns the active fixture manager instance"""

    # ---------------------------------- PROTECTED METHODS -------------------------------------------------------------

    @abstractmethod
    def _prepare_execution(self, show_discarded):
        """
        This method runs before the branch will be executed and before the fixture construction code of this branch
        runs.
        """

    @abstractmethod
    def _body_execution(self, show_discarded):
        """
        This method runs between the fixture construction and teardown code. It should trigger the execution of the
        child branches.
        """

    @abstractmethod
    def _cleanup_execution(self, show_discarded):
        """
        This method runs after the branch was executed (also after the fixture teardown code ran)
        """

    # ---------------------------------- METHODS -----------------------------------------------------------------------

    def set_result_for_whole_branch(self, value: ResultState):
        """
        This method sets the executor result for all sub executors.

        :param value: the new value that should be set for this branch
        """
        if value not in (ResultState.SKIP, ResultState.COVERED_BY, ResultState.NOT_RUN):
            raise ValueError("can not set a state that is not NOT_RUN, SKIP or COVERED_BY for a whole branch")
        if self.all_child_executors is None:
            self.body_result.set_result(result=value, exception=None)
        else:
            for cur_child_executor in self.all_child_executors:
                cur_child_executor.set_result_for_whole_branch(value)

    @abstractmethod
    def cleanup_empty_executor_branches(self, consider_discarded=False):
        """
        This method searches the whole tree and removes branches where an executor item has no own children. It can
        remove these branches, because they have no valid matchings.

        :param consider_discarded: true if this method should consider discarded branches, otherwise False
        """

    def filter_tree_for_user_filters(self):
        """
        This method calls all user defined filters that are to be applied to the executor tree.
        """
        if self.all_child_executors:
            for cur_child_executor in self.all_child_executors:
                cur_child_executor.filter_tree_for_user_filters()

    def execute(self, show_discarded=False):
        """
        Executes the whole branch
        """
        start_time = time.perf_counter()
        self._prepare_execution(show_discarded=show_discarded)

        try:
            try:
                if self.has_runnable_tests():
                    self.fixture_manager.enter(self)
                    self.construct_result.set_result(ResultState.SUCCESS)
                else:
                    self.construct_result.set_result(ResultState.NOT_RUN)

                self._body_execution(show_discarded=show_discarded)
            except Exception as exc:  # pylint: disable=broad-exception-caught
                # this has to be a construction fixture error
                traceback.print_exception(*sys.exc_info())
                self.construct_result.set_result(ResultState.ERROR, exc)
            finally:
                if self.has_runnable_tests():
                    if self.fixture_manager.is_allowed_to_leave(self):
                        self.fixture_manager.leave(self)
                        self.teardown_result.set_result(ResultState.SUCCESS)
                else:
                    self.teardown_result.set_result(ResultState.NOT_RUN)
        except Exception as exc:  # pylint: disable=broad-exception-caught
            # this has to be a teardown fixture error
            traceback.print_exception(*sys.exc_info())
            self.teardown_result.set_result(ResultState.ERROR, exc)

        self._cleanup_execution(show_discarded=show_discarded)

        self.execution_time_sec = time.perf_counter() - start_time



from ...lib.features import FeatureIII, FeatureIV
from ...balderglob import RuntimeObserver


class SetupFeatureIV(FeatureIV):
    feature_iii = FeatureIII()

    def do_something(self):
        RuntimeObserver.add_entry(
            __file__, SetupFeatureIV, SetupFeatureIV.do_something, "enter `SetupFeatureIV.do_something`",
            category="feature")
        self.feature_iii.called_from_outer_feature()

    def called_from_outer_feature(self):
        RuntimeObserver.add_entry(
            __file__, SetupFeatureIV, SetupFeatureIV.called_from_outer_feature,
            "enter `SetupFeatureIV.called_from_outer_feature`", category="feature")



from .feature_i import SetupFeatureI
from .feature_ii import SetupFeatureII
from .newly_setup_feature import NewlySetupFeature

__all__ = [
    "SetupFeatureI",
    "SetupFeatureII",
    "NewlySetupFeature"
]



from __future__ import annotations
from typing import List, Union, TYPE_CHECKING

from .base_connection_relation import BaseConnectionRelation, BaseConnectionRelationT

if TYPE_CHECKING:
    from ..connection import Connection


class OrConnectionRelation(BaseConnectionRelation):
    """
    describes a OR relation between connections
    """

    def get_tree_str(self) -> str:
        based_on_strings = [cur_elem.get_tree_str() for cur_elem in self._connections]
        return f"({' & '.join(based_on_strings)})"

    def get_simplified_relation(self) -> OrConnectionRelation:
        from ..connection import Connection  # pylint: disable=import-outside-toplevel

        result = OrConnectionRelation()
        for cur_inner_elem in self.connections:
            if isinstance(cur_inner_elem, Connection):
                # that is fine - just add it again
                result.append(cur_inner_elem)
            elif isinstance(cur_inner_elem, BaseConnectionRelation):
                # simplify this AND/OR relation and add the items of it
                # (the simplified version is always an OR relation!)
                result.extend(cur_inner_elem.get_simplified_relation())
            else:
                raise TypeError(f'detect unexpected element type `{cur_inner_elem.__class__}` in inner elements')
        return result

    def is_single(self) -> bool:
        if len(self.connections) == 0:
            return True
        if len(self.connections) == 1:
            return self.connections[0].is_single()
        return False

    def get_singles(self) -> List[Connection]:
        result = []
        for elem in self.connections:
            result.extend(elem.get_singles())
        return result

    def cut_into_all_possible_subtree_branches(self) -> List[OrConnectionRelation]:
        if not self.is_single():
            raise ValueError('can not execute method, because relation is not single')
        result = [OrConnectionRelation()]
        if len(self.connections) == 0:
            return result
        result.extend(self.connections[0].cut_into_all_possible_subtree_branches())
        return result

    def contained_in(self, other_conn: Union[Connection, BaseConnectionRelationT], ignore_metadata=False) -> bool:
        if not self.is_resolved():
            raise ValueError('can not execute method, because connection relation is not resolved')
        if not other_conn.is_resolved():
            raise ValueError('can not execute method, because other connection relation is not resolved')
        for cur_inner_cnn in self._connections:
            if cur_inner_cnn.contained_in(other_conn, ignore_metadata=ignore_metadata):
                return True
        return False



from __future__ import annotations


class BalderSettings:
    """
    This class can be overwritten to manipulate the default settings for the balder test system. You can overwrite these
    settings by defining a subclass in your `balderglob.py`.
    """

    #: specifies that the test run should include duplicated tests that are declared as ``@covered_by`` another test
    #:   method
    force_covered_by_duplicates = False

    #: specifies the connection tree identifier that should be used as global identifier ("" is the default one)
    used_global_connection_tree = ""



from __future__ import annotations

from dataclasses import dataclass, fields
from typing import TYPE_CHECKING, List, Union

from enum import Enum

if TYPE_CHECKING:
    from _balder.executor.basic_executor import BasicExecutor


class ResultState(Enum):
    """
    enumeration that describes the possible results of a testcase-/fixture-executor
    """
    # this state will be assigned if the executor doesn't run yet
    NOT_RUN = 'not_run'
    # this state will be assigned if the test fails (only assignable to :class:`TestcaseExecutor`)
    FAILURE = 'failure'
    # this state will be assigned if the executor can not be executed because the fixture fails (only possible in
    # construction part of fixtures, if the error occurs in teardown the next higher executer get this state) - will be
    # assigned to the executor which has the fixture that fails
    ERROR = 'error'
    # this state will be assigned if the executor was executed successfully (session fixture and teardown fixture; for
    # :class:`TestcaseExecutor` also the testcase itself)
    SUCCESS = 'success'
    # this state will be assigned if the executor was skipped
    SKIP = 'skip'
    # this state will be assigned if the executor is covered by another executor and wasn't executed (only assignable
    # to :class:`ScenarioExecutor` and :class:`TestcaseExecutor`)
    COVERED_BY = 'covered_by'

    @staticmethod
    def priority_order() -> List[ResultState]:
        """
        returns an order of states that returns the highest priority as first item and the lowest as last one
        """
        return [ResultState.ERROR, ResultState.FAILURE, ResultState.SUCCESS, ResultState.COVERED_BY, ResultState.SKIP,
                ResultState.NOT_RUN]


class _Result:

    def __init__(self, executor: BasicExecutor):
        self._result = ResultState.NOT_RUN

        self._related_executor = executor
        #: contains the exception that was thrown and returned into this result
        self.exception = None

        self.char_mapping = {
            ResultState.NOT_RUN: " ",
            ResultState.FAILURE: "X",
            ResultState.ERROR: "E",
            ResultState.SUCCESS: ".",
            ResultState.SKIP: "S",
            ResultState.COVERED_BY: "C"
        }

    def __str__(self):
        return self.result.name

    @property
    def result(self) -> ResultState:
        """returns the result"""
        return self._result

    def get_result_as_char(self):
        """
        returns a single char that represents the result state
        """

        return self.char_mapping[self.result]


class _SettableResult(_Result):
    ALLOWED_STATES = [ResultState.NOT_RUN, ResultState.FAILURE, ResultState.ERROR, ResultState.SUCCESS,
                      ResultState.SKIP, ResultState.COVERED_BY]

    def set_result(self, result: ResultState, exception: Union[Exception, None] = None) -> None:
        """this method allows to set the result - if the value is SKIP or COVERED_BY this method automatically sets this
        values also for all components of the branch in related executor"""

        if result not in self.ALLOWED_STATES:
            raise ValueError(
                f"it is not allowed to set the given result `{result.name}` in a ´{self.__class__.__name__}´")
        self._result = result
        self.exception = exception


class BranchBodyResult(_Result):
    """
    This result is used in every branch body. So it isn't set directly, it is always determined depending on the
    :class:`FixturePartResult` and :class:`TestcaseResult` of the executors children.
    """

    def __init__(self, executor: BasicExecutor):
        from _balder.executor.testcase_executor import TestcaseExecutor  # pylint: disable=import-outside-toplevel

        if isinstance(executor, TestcaseExecutor):
            raise TypeError("testcase executors are not allowed to use in `BranchBodyResult`")
        super().__init__(executor=executor)

    @property
    def result(self):
        """returns the determined result for the inner branch of the related executor"""
        relative_result = ResultState.NOT_RUN
        priority_order = ResultState.priority_order()
        if self._related_executor.all_child_executors:
            for cur_child in self._related_executor.all_child_executors:
                if priority_order.index(cur_child.executor_result) < priority_order.index(relative_result):
                    relative_result = cur_child.executor_result
        self._result = relative_result
        return self._result


class FixturePartResult(_SettableResult):
    """
    This result is used for fixture construct or teardown code. It describes if there was an error in the construct
    or teardown part of the fixture code for an executor.
    """
    #: contains the possible values that can be set for this Result type
    ALLOWED_STATES = [ResultState.NOT_RUN, ResultState.ERROR, ResultState.SUCCESS]


class TestcaseResult(_SettableResult):
    """
    This result is used for real executed testcase methods. These elements can only be assigned to real testcase methods
    which are only available in :class:`TestcaseExecutor`.
    """
    #: contains the possible values that can be set for this Result type
    ALLOWED_STATES = [ResultState.NOT_RUN, ResultState.FAILURE, ResultState.SUCCESS, ResultState.SKIP,
                      ResultState.COVERED_BY]

@dataclass
class ResultSummary:
    """
    object that holds the test results for multiple tests like on branch or global level
    """
    # this state will be assigned if the executor doesn't run yet
    not_run: int = 0
    # this state will be assigned if the test fails (only assignable to :class:`TestcaseExecutor`)
    failure: int = 0
    # this state will be assigned if the executor can not be executed because the fixture fails (only possible in
    # construction part of fixtures, if the error occurs in teardown the next higher executer get this state) - will be
    # assigned to the executor which has the fixture that fails
    error: int = 0
    # this state will be assigned if the executor was executed successfully (session fixture and teardown fixture; for
    # :class:`TestcaseExecutor` also the testcase itself)
    success: int = 0
    # this state will be assigned if the executor was skipped
    skip: int = 0
    # this state will be assigned if the executor is covered by another executor and wasn't executed (only assignable
    # to :class:`ScenarioExecutor` and :class:`TestcaseExecutor`)
    covered_by: int = 0

    def __add__(self, other) -> ResultSummary:
        new_summary = ResultSummary()
        for cur_field in fields(self.__class__):
            self_val = getattr(self, cur_field.name)
            other_val = getattr(other, cur_field.name)
            setattr(new_summary, cur_field.name, self_val + other_val)
        return new_summary



from ...lib.features import FeatureOfRelevanceLvl2
from ...balderglob import RuntimeObserver


class SetupFeatureOfRelevanceLvl3(FeatureOfRelevanceLvl2):

    def do_something(self):
        RuntimeObserver.add_entry(
            __file__, SetupFeatureOfRelevanceLvl3, SetupFeatureOfRelevanceLvl3.do_something,
            "enter `SetupFeatureOfRelevanceLvl3.do_something`", category="feature")



from multiprocessing import Process
from _balder.balder_session import BalderSession
from _balder.testresult import ResultState


def test_0_use_default_globtree(balder_working_dir):
    """
    This tests secures that the default globtree is used if there is no special setting done in project's
    `BalderSettings` object.
    """
    proc = Process(target=processed, args=(balder_working_dir,))
    proc.start()
    proc.join()
    assert proc.exitcode == 0, "the process terminates with an error"


def processed(env_dir):

    print("\n", flush=True)
    session = BalderSession(cmd_args=[], working_dir=env_dir)
    session.run()
    print("\n")

    # check result states everywhere (have to be SUCCESS everywhere)
    assert session.executor_tree.executor_result == ResultState.SUCCESS, "test session does not terminates with SUCCESS"
    # check that there are all expected mappings
    assert len(session.executor_tree.get_all_testcase_executors()) == 4, "do not resolve to exactly 4 testcases"



from _balder.testresult import ResultState
from _balder.balder_session import BalderSession
from ...test_utilities.base_0_envtester_class import Base0EnvtesterClass


class Test0OnlyScenarioB(Base0EnvtesterClass):
    """
    This testcase executes the basic example and checks if the tree ends with the result SUCCESS. It calls the balder
    environment with the following filter patterns for setup and for scenarios and validates that the correct tests are
    executed:

    `balder .. --only-with-scenario scenarios/*_b.py`

    Only the elements matching with `ScenarioB` should be executed here.
    """
    @property
    def cmd_args(self):
        return ["--only-with-scenario", "scenarios/*_b.py"]

    @property
    def expected_data(self):
        return (
            # FIXTURE-CONSTRUCTION: balderglob_fixture_session
            {"file": "balderglob.py", "meth": "balderglob_fixture_session", "part": "construction"},
            # FIXTURE-CONSTRUCTION: SetupB.fixture_session
            {"cls": "SetupB", "meth": "fixture_session", "part": "construction"},
            {"cls": "SetupFeatureIII", "meth": "do_something", "category": "feature"},
            {"cls": "SetupFeatureIV", "meth": "do_something", "category": "feature"},
            # FIXTURE-CONSTRUCTION: ScenarioB.fixture_session
            {"cls": "ScenarioB", "meth": "fixture_session", "part": "construction"},
            {"cls": "FeatureIII", "meth": "do_something", "category": "feature"},
            {"cls": "FeatureIV", "meth": "do_something", "category": "feature"},
            # FIXTURE-CONSTRUCTION: balderglob_fixture_setup
            {"file": "balderglob.py", "meth": "balderglob_fixture_setup", "part": "construction"},
            # FIXTURE-CONSTRUCTION: SetupB.fixture_setup
            {"cls": "SetupB", "meth": "fixture_setup", "part": "construction"},
            {"cls": "SetupFeatureIII", "meth": "do_something", "category": "feature"},
            {"cls": "SetupFeatureIV", "meth": "do_something", "category": "feature"},
            # FIXTURE-CONSTRUCTION: ScenarioB.fixture_setup
            {"cls": "ScenarioB", "meth": "fixture_setup", "part": "construction"},
            {"cls": "FeatureIII", "meth": "do_something", "category": "feature"},
            {"cls": "FeatureIV", "meth": "do_something", "category": "feature"},
            # FIXTURE-CONSTRUCTION: balderglob_fixture_scenario
            {"file": "balderglob.py", "meth": "balderglob_fixture_scenario", "part": "construction"},
            # FIXTURE-CONSTRUCTION: SetupB.fixture_scenario
            {"cls": "SetupB", "meth": "fixture_scenario", "part": "construction"},
            {"cls": "SetupFeatureIII", "meth": "do_something", "category": "feature"},
            {"cls": "SetupFeatureIV", "meth": "do_something", "category": "feature"},
            # FIXTURE-CONSTRUCTION: ScenarioB.fixture_scenario
            {"cls": "ScenarioB", "meth": "fixture_scenario", "part": "construction"},
            {"cls": "FeatureIII", "meth": "do_something", "category": "feature"},
            {"cls": "FeatureIV", "meth": "do_something", "category": "feature"},
            # FIXTURE-CONSTRUCTION: balderglob_fixture_variation
            {"file": "balderglob.py", "meth": "balderglob_fixture_variation", "part": "construction"},
            # FIXTURE-CONSTRUCTION: SetupB.fixture_variation
            {"cls": "SetupB", "meth": "fixture_variation", "part": "construction"},
            {"cls": "SetupFeatureIII", "meth": "do_something", "category": "feature"},
            {"cls": "SetupFeatureIV", "meth": "do_something", "category": "feature"},
            # FIXTURE-CONSTRUCTION: ScenarioB.fixture_variation
            {"cls": "ScenarioB", "meth": "fixture_variation", "part": "construction"},
            {"cls": "SetupFeatureIII", "meth": "do_something", "category": "feature"},
            {"cls": "SetupFeatureIV", "meth": "do_something", "category": "feature"},
            [
                (
                    # FIXTURE-CONSTRUCTION: balderglob_fixture_testcase
                    {"file": "balderglob.py", "meth": "balderglob_fixture_testcase", "part": "construction"},
                    # FIXTURE-CONSTRUCTION: SetupB.fixture_testcase
                    {"cls": "SetupB", "meth": "fixture_testcase", "part": "construction"},
                    {"cls": "SetupFeatureIII", "meth": "do_something", "category": "feature"},
                    {"cls": "SetupFeatureIV", "meth": "do_something", "category": "feature"},
                    # FIXTURE-CONSTRUCTION: ScenarioB.fixture_testcase
                    {"cls": "ScenarioB", "meth": "fixture_testcase", "part": "construction"},
                    {"cls": "SetupFeatureIII", "meth": "do_something", "category": "feature"},
                    {"cls": "SetupFeatureIV", "meth": "do_something", "category": "feature"},
                    # TESTCASE: ScenarioA.test_b_1
                    {"cls": "ScenarioB", "meth": "test_b_1"},
                    {"cls": "SetupFeatureIII", "meth": "do_something", "category": "feature"},
                    {"cls": "SetupFeatureIV", "meth": "do_something", "category": "feature"},
                    # FIXTURE-TEARDOWN: ScenarioB.fixture_testcase
                    {"cls": "ScenarioB", "meth": "fixture_testcase", "part": "teardown"},
                    {"cls": "SetupFeatureIII", "meth": "do_something", "category": "feature"},
                    {"cls": "SetupFeatureIV", "meth": "do_something", "category": "feature"},
                    # FIXTURE-TEARDOWN: SetupB.fixture_testcase
                    {"cls": "SetupB", "meth": "fixture_testcase", "part": "teardown"},
                    {"cls": "SetupFeatureIII", "meth": "do_something", "category": "feature"},
                    {"cls": "SetupFeatureIV", "meth": "do_something", "category": "feature"},
                    # FIXTURE-TEARDOWN: balderglob_fixture_testcase
                    {"file": "balderglob.py", "meth": "balderglob_fixture_testcase", "part": "teardown"},
                ),
                (
                    # FIXTURE-CONSTRUCTION: balderglob_fixture_testcase
                    {"file": "balderglob.py", "meth": "balderglob_fixture_testcase", "part": "construction"},
                    # FIXTURE-CONSTRUCTION: SetupB.fixture_testcase
                    {"cls": "SetupB", "meth": "fixture_testcase", "part": "construction"},
                    {"cls": "SetupFeatureIII", "meth": "do_something", "category": "feature"},
                    {"cls": "SetupFeatureIV", "meth": "do_something", "category": "feature"},
                    # FIXTURE-CONSTRUCTION: ScenarioB.fixture_testcase
                    {"cls": "ScenarioB", "meth": "fixture_testcase", "part": "construction"},
                    {"cls": "SetupFeatureIII", "meth": "do_something", "category": "feature"},
                    {"cls": "SetupFeatureIV", "meth": "do_something", "category": "feature"},
                    # TESTCASE: ScenarioA.test_b_2
                    {"cls": "ScenarioB", "meth": "test_b_2"},
                    {"cls": "SetupFeatureIII", "meth": "do_something", "category": "feature"},
                    {"cls": "SetupFeatureIV", "meth": "do_something", "category": "feature"},
                    # FIXTURE-TEARDOWN: ScenarioB.fixture_testcase
                    {"cls": "ScenarioB", "meth": "fixture_testcase", "part": "teardown"},
                    {"cls": "SetupFeatureIII", "meth": "do_something", "category": "feature"},
                    {"cls": "SetupFeatureIV", "meth": "do_something", "category": "feature"},
                    # FIXTURE-TEARDOWN: SetupB.fixture_testcase
                    {"cls": "SetupB", "meth": "fixture_testcase", "part": "teardown"},
                    {"cls": "SetupFeatureIII", "meth": "do_something", "category": "feature"},
                    {"cls": "SetupFeatureIV", "meth": "do_something", "category": "feature"},
                    # FIXTURE-TEARDOWN: balderglob_fixture_testcase
                    {"file": "balderglob.py", "meth": "balderglob_fixture_testcase", "part": "teardown"},
                ),
            ],
            # FIXTURE-TEARDOWN: ScenarioB.fixture_variation
            {"cls": "ScenarioB", "meth": "fixture_variation", "part": "teardown"},
            {"cls": "SetupFeatureIII", "meth": "do_something", "category": "feature"},
            {"cls": "SetupFeatureIV", "meth": "do_something", "category": "feature"},
            # FIXTURE-TEARDOWN: SetupB.fixture_variation
            {"cls": "SetupB", "meth": "fixture_variation", "part": "teardown"},
            {"cls": "SetupFeatureIII", "meth": "do_something", "category": "feature"},
            {"cls": "SetupFeatureIV", "meth": "do_something", "category": "feature"},
            # FIXTURE-TEARDOWN: balderglob_fixture_variation
            {"file": "balderglob.py", "meth": "balderglob_fixture_variation", "part": "teardown"},
            # FIXTURE-TEARDOWN: ScenarioB.fixture_scenario
            {"cls": "ScenarioB", "meth": "fixture_scenario", "part": "teardown"},
            {"cls": "FeatureIII", "meth": "do_something", "category": "feature"},
            {"cls": "FeatureIV", "meth": "do_something", "category": "feature"},
            # FIXTURE-TEARDOWN: SetupB.fixture_scenario
            {"cls": "SetupB", "meth": "fixture_scenario", "part": "teardown"},
            {"cls": "SetupFeatureIII", "meth": "do_something", "category": "feature"},
            {"cls": "SetupFeatureIV", "meth": "do_something", "category": "feature"},
            # FIXTURE-TEARDOWN: balderglob_fixture_scenario
            {"file": "balderglob.py", "meth": "balderglob_fixture_scenario", "part": "teardown"},
            # FIXTURE-TEARDOWN: ScenarioB.fixture_setup
            {"cls": "ScenarioB", "meth": "fixture_setup", "part": "teardown"},
            {"cls": "FeatureIII", "meth": "do_something", "category": "feature"},
            {"cls": "FeatureIV", "meth": "do_something", "category": "feature"},
            # FIXTURE-TEARDOWN: SetupB.fixture_setup
            {"cls": "SetupB", "meth": "fixture_setup", "part": "teardown"},
            {"cls": "SetupFeatureIII", "meth": "do_something", "category": "feature"},
            {"cls": "SetupFeatureIV", "meth": "do_something", "category": "feature"},
            # FIXTURE-TEARDOWN: balderglob_fixture_setup
            {"file": "balderglob.py", "meth": "balderglob_fixture_setup", "part": "teardown"},
            # FIXTURE-TEARDOWN: ScenarioB.fixture_session
            {"cls": "ScenarioB", "meth": "fixture_session", "part": "teardown"},
            {"cls": "FeatureIII", "meth": "do_something", "category": "feature"},
            {"cls": "FeatureIV", "meth": "do_something", "category": "feature"},
            # FIXTURE-TEARDOWN: SetupA.fixture_session
            {"cls": "SetupB", "meth": "fixture_session", "part": "teardown"},
            {"cls": "SetupFeatureIII", "meth": "do_something", "category": "feature"},
            {"cls": "SetupFeatureIV", "meth": "do_something", "category": "feature"},
            # FIXTURE-TEARDOWN: balderglob_fixture_session
            {"file": "balderglob.py", "meth": "balderglob_fixture_session", "part": "teardown"},
        )

    @staticmethod
    def validate_finished_session(session: BalderSession):
        # check result states everywhere (have to be SUCCESS everywhere
        assert session.executor_tree.executor_result == ResultState.SUCCESS, \
            "test session does not terminates with success"

        assert session.executor_tree.construct_result.result == ResultState.SUCCESS, \
            "global executor tree construct part does not set ResultState.SUCCESS"
        assert session.executor_tree.body_result.result == ResultState.SUCCESS, \
            "global executor tree body part does not set ResultState.SUCCESS"
        assert session.executor_tree.teardown_result.result == ResultState.SUCCESS, \
            "global executor tree teardown part does not set ResultState.SUCCESS"
        for cur_setup_executor in session.executor_tree.get_setup_executors():
            assert cur_setup_executor.executor_result == ResultState.SUCCESS, \
                "the setup executor does not have result SUCCESS"

            assert cur_setup_executor.construct_result.result == ResultState.SUCCESS
            assert cur_setup_executor.body_result.result == ResultState.SUCCESS
            assert cur_setup_executor.teardown_result.result == ResultState.SUCCESS

            for cur_scenario_executor in cur_setup_executor.get_scenario_executors():
                assert cur_scenario_executor.executor_result == ResultState.SUCCESS, \
                    "the scenario executor does not have result SUCCESS"

                assert cur_scenario_executor.construct_result.result == ResultState.SUCCESS
                assert cur_scenario_executor.body_result.result == ResultState.SUCCESS
                assert cur_scenario_executor.teardown_result.result == ResultState.SUCCESS

                for cur_variation_executor in cur_scenario_executor.get_variation_executors():
                    assert cur_variation_executor.executor_result == ResultState.SUCCESS, \
                        "the variation executor does not have result SUCCESS"

                    assert cur_variation_executor.construct_result.result == ResultState.SUCCESS
                    assert cur_variation_executor.body_result.result == ResultState.SUCCESS
                    assert cur_variation_executor.teardown_result.result == ResultState.SUCCESS

                    for cur_testcase_executor in cur_variation_executor.get_testcase_executors():
                        assert cur_testcase_executor.executor_result == ResultState.SUCCESS, \
                            "the testcase executor does not have result SUCCESS"

                        assert cur_testcase_executor.construct_result.result == ResultState.SUCCESS
                        assert cur_testcase_executor.body_result.result == ResultState.SUCCESS
                        assert cur_testcase_executor.teardown_result.result == ResultState.SUCCESS



from .scenario_b_parent import ScenarioBParent
from ..balderglob import RuntimeObserver


class ScenarioBSub2(ScenarioBParent):
    IGNORE = ['test_b_1']
    SKIP = ['test_b_2']

    def test_b_3(self):
        RuntimeObserver.add_entry(__file__, ScenarioBSub2, ScenarioBSub2.test_b_3, category="testcase",
                                  msg=f"execute Test `{ScenarioBSub2.test_b_3.__qualname__}`")
        self.ScenarioDevice3.iii.do_something()
        self.ScenarioDevice4.iv.do_something()



import balder
from ...balderglob import RuntimeObserver


class SetupFeatureNew(balder.Feature):

    def do_something(self):
        RuntimeObserver.add_entry(
            __file__, SetupFeatureNew, SetupFeatureNew.do_something, "enter `SetupFeatureNew.do_something`",
            category="feature")



import balder
from ..balderglob import RuntimeObserver


class FeatureI(balder.Feature):

    def do_something(self):
        RuntimeObserver.add_entry(
            __file__, FeatureI, FeatureI.do_something, "enter `FeatureI.do_something`", category="feature")

    def called_from_outer_feature(self):
        RuntimeObserver.add_entry(
            __file__, FeatureI, FeatureI.called_from_outer_feature, "enter `FeatureI.called_from_outer_feature`",
            category="feature")


class FeatureII(balder.Feature):
    def do_something(self):
        RuntimeObserver.add_entry(
            __file__, FeatureII, FeatureII.do_something, "enter `FeatureII.do_something`", category="feature")

    def called_from_outer_feature(self):
        RuntimeObserver.add_entry(
            __file__, FeatureI, FeatureI.called_from_outer_feature, "enter `FeatureII.called_from_outer_feature`",
            category="feature")


class FeatureIII(balder.Feature):

    def do_something(self):
        RuntimeObserver.add_entry(
            __file__, FeatureIII, FeatureIII.do_something, "enter `FeatureIII.do_something`", category="feature")

    def called_from_outer_feature(self):
        RuntimeObserver.add_entry(
            __file__, FeatureIV, FeatureIV.called_from_outer_feature, "enter `FeatureIII.called_from_outer_feature`",
            category="feature")


class FeatureIV(balder.Feature):

    def do_something(self):
        RuntimeObserver.add_entry(
            __file__, FeatureIV, FeatureIV.do_something, "enter `FeatureIV.do_something`", category="feature")

    def called_from_outer_feature(self):
        RuntimeObserver.add_entry(
            __file__, FeatureIV, FeatureIV.called_from_outer_feature, "enter `FeatureIV.called_from_outer_feature`",
            category="feature")


class FeatureV(balder.Feature):

    def do_something(self):
        RuntimeObserver.add_entry(
            __file__, FeatureV, FeatureV.do_something, "enter `FeatureV.do_something`", category="feature")

    def called_from_outer_feature(self):
        RuntimeObserver.add_entry(
            __file__, FeatureV, FeatureV.called_from_outer_feature, "enter `FeatureV.called_from_outer_feature`",
            category="feature")


class FeatureVI(balder.Feature):
    def do_something(self):
        RuntimeObserver.add_entry(
            __file__, FeatureVI, FeatureVI.do_something, "enter `FeatureVI.do_something`", category="feature")

    def called_from_outer_feature(self):
        RuntimeObserver.add_entry(
            __file__, FeatureVI, FeatureVI.called_from_outer_feature, "enter `FeatureVI.called_from_outer_feature`",
            category="feature")


class FeatureVII(balder.Feature):

    def do_something(self):
        RuntimeObserver.add_entry(
            __file__, FeatureVII, FeatureVII.do_something, "enter `FeatureVII.do_something`", category="feature")

    def called_from_outer_feature(self):
        RuntimeObserver.add_entry(
            __file__, FeatureVII, FeatureVII.called_from_outer_feature,
            "enter `FeatureVII.called_from_outer_feature`", category="feature")


class FeatureVIII(balder.Feature):

    def do_something(self):
        RuntimeObserver.add_entry(
            __file__, FeatureVIII, FeatureVIII.do_something,
            "enter `FeatureVIII.do_something`", category="feature")

    def called_from_outer_feature(self):
        RuntimeObserver.add_entry(
            __file__, FeatureVIII, FeatureVIII.called_from_outer_feature,
            "enter `FeatureVIII.called_from_outer_feature`", category="feature")



from _balder.vdevice import VDevice
from _balder.exceptions import AccessToUnmappedVDeviceException


class UnmappedVDevice(VDevice):
    """
    This special vdevice class will be assigned to all existing VDevice's that are not mapped during the execution of
    a variation.
    """

    def __getattr__(self, item):
        raise AccessToUnmappedVDeviceException('it is not allowed to access the attributes of an unmapped VDevice - '
                                               'did you forget to map it?')



from multiprocessing import Process

from _balder.testresult import ResultState
from _balder.balder_session import BalderSession


def test_1_vdevice_which_is_no_part_of_a_variation(balder_working_dir):
    """
    This testcase is a modified version of the calculator environment.

    The setup defines two calculator devices `Calculator1` and `Calculator2` that theoretically could map to the
    scenario device `Calculator`. Normally both of these setup devices `Calculator1` and `Calculator2` should map.

    However, since we have a VDevice mapping on setup level only to the `Calculator1` device, balder should not allow
    the second variation, at which the `Calculator2` would be mapped to the scenario device `Calculator`, because it
    cannot be used as VDevice.

    The test secures that only the mapped `Calculator1` is a valid variation.
    """
    proc = Process(target=processed, args=(balder_working_dir, ))
    proc.start()
    proc.join()
    assert proc.exitcode == 0, "the process terminates with an error"


def processed(env_dir):
    print("\n", flush=True)
    session = BalderSession(cmd_args=[], working_dir=env_dir)
    session.run()
    print("\n")

    assert session.executor_tree.executor_result == ResultState.SUCCESS, \
        "test session does not terminates with success"
    assert len(session.executor_tree.get_setup_executors()) == 1

    scenario_executors = session.executor_tree.get_setup_executors()[0].get_scenario_executors()
    assert len(scenario_executors) == 1
    assert len(scenario_executors[0].get_variation_executors()) == 2
    for cur_variation_executor in scenario_executors[0].get_variation_executors():
        all_setup_devices = [cur_device.__name__ for cur_device in cur_variation_executor.base_device_mapping.values()]
        assert "Calculator1" in all_setup_devices, \
            "can not find the expected `Calculator` setup device in mapped-devices"
        assert "Calculator2" not in all_setup_devices, \
            "`Calculator2` should not be contained in mapped-devices"



from __future__ import annotations
from typing import Type, List, Union, Optional, Dict, TYPE_CHECKING

import logging
import inspect
from abc import ABC, abstractmethod
from _balder.setup import Setup
from _balder.device import Device
from _balder.scenario import Scenario
from _balder.connection_metadata import ConnectionMetadata
from _balder.controllers.controller import Controller
from _balder.controllers.device_controller import DeviceController
from _balder.controllers.vdevice_controller import VDeviceController
from _balder.exceptions import MultiInheritanceError, DeviceOverwritingError, MissingFeaturesOfVDeviceError

if TYPE_CHECKING:
    from _balder.connection import Connection
    from _balder.controllers import ScenarioController
    from _balder.controllers import SetupController

logger = logging.getLogger(__file__)


class NormalScenarioSetupController(Controller, ABC):
    """
    This is the abstract base controller class for the Scenario- and Setup-Controller.
    """
    # describes if the current controller is for setups or for scenarios (has to be set in child controller)
    _related_type: Optional[Type[Scenario, Setup]] = None

    # ---------------------------------- STATIC METHODS ----------------------------------------------------------------

    # ---------------------------------- CLASS METHODS -----------------------------------------------------------------

    # ---------------------------------- PROPERTIES --------------------------------------------------------------------

    # ---------------------------------- PROTECTED METHODS -------------------------------------------------------------

    # ---------------------------------- METHODS -----------------------------------------------------------------------

    @staticmethod
    def get_for(related_cls) -> Union[ScenarioController, SetupController]:
        """
        This class returns the current existing controller instance for the given item. If the instance does not exist
        yet, it will automatically create it and saves the instance in an internal dictionary.

        .. note::
            This method automatically returns the correct controller type, depending on the class you provide with
            `related_cls`.
        """
        # pylint: disable-next=import-outside-toplevel
        from _balder.controllers.setup_controller import SetupController
        # pylint: disable-next=import-outside-toplevel
        from _balder.controllers.scenario_controller import ScenarioController

        if issubclass(related_cls, Scenario):
            return ScenarioController.get_for(related_cls)
        if issubclass(related_cls, Setup):
            return SetupController.get_for(related_cls)

        raise TypeError(f"illegal non supported type `{related_cls.__name__}` given for `related_cls`")

    def get_all_inner_device_classes(self) -> List[Type[Device]]:
        """
        This method provides a list of all :meth:`Device` classes that have been defined as inner classes in the related
        scenario or setup.
        """

        all_classes = inspect.getmembers(self.related_cls, inspect.isclass)
        filtered_classes = []
        for _, cur_class in all_classes:
            if not issubclass(cur_class, Device):
                # filter all classes and make sure that only the child classes of :meth:`Device` remain
                continue
            if DeviceController.get_for(cur_class).get_outer_class() != self.related_cls:
                # filter all classes that do not match the setup name in __qualname__
                continue
            # otherwise, add this candidate
            filtered_classes.append(cur_class)
        return filtered_classes

    def get_inner_device_class_by_string(self, device_str: str) -> Union[Type[Device], None]:
        """
        This method returns the inner Device class for the given string.

        :param device_str: the name string of the Device that should be returned

        :return: the Device class or None, if the method has not found any class with this name
        """
        possible_devs = [cur_vdevice for cur_vdevice in self.get_all_inner_device_classes()
                         if cur_vdevice.__name__ == device_str]
        if len(possible_devs) == 0:
            return None
        if len(possible_devs) > 1:
            raise RuntimeError("found more than one possible vDevices - something unexpected happened")

        return possible_devs[0]

    def get_all_abs_inner_device_classes(self) -> List[Type[Device]]:
        """
        This method provides a list of all :meth:`Device` classes that are valid for the related scenario or setup
        class. If the class itself does not implement some inner devices by its own, it returns the absolute inner
        devices of the next higher setup/scenario parent class.
        """
        cls_devices = self.get_all_inner_device_classes()
        if len(cls_devices) == 0:
            # search for parent class
            base_class = self.get_next_parent_class()

            if base_class is None:
                # if the class type is the original `Setup` or `Scenario` type -> no inner devices exists
                return []
            return self.__class__.get_for(base_class).get_all_abs_inner_device_classes()

        return cls_devices

    def get_abs_inner_device_class_by_string(self, device_str: str) -> Union[Type[Device], None]:
        """
        This method returns the absolute inner Device class for the given string.

        :param device_str: the name string of the Device that should be returned

        :return: the Device class or None, if the method has not found any class with this name
        """
        possible_devs = [cur_vdevice for cur_vdevice in self.get_all_abs_inner_device_classes()
                         if cur_vdevice.__name__ == device_str]
        if len(possible_devs) == 0:
            return None
        if len(possible_devs) > 1:
            raise RuntimeError("found more than one possible vDevices - something unexpected happened")

        return possible_devs[0]

    @abstractmethod
    def get_next_parent_class(self) -> Union[Type[Scenario], Type[Setup], None]:
        """
        This method returns the next parent class which is a subclass of the :class:`Scenario`/:class:`Setup` itself.

        :return: returns the next parent class or None if the next parent class is :class:`Scenario`/:class:`Setup`
                 itself
        """

    def get_all_connections(self) -> List[Connection]:
        """
        This method returns all connection objects which have been defined between devices of the related class.
        """
        all_device_classes = self.get_all_inner_device_classes()
        all_connections = []
        for cur_device in all_device_classes:
            cur_device_controller = DeviceController.get_for(cur_device)
            for cur_node_name in cur_device_controller.connections.keys():
                all_connections += cur_device_controller.connections[cur_node_name]
        return all_connections

    def get_all_abs_connections(self) -> List[Connection]:
        """
        This method returns all absolute connection objects which have been defined between absolute devices of the
        related scenario or setup class.
        """
        all_device_classes = self.get_all_abs_inner_device_classes()
        all_connections = []
        for cur_device in all_device_classes:
            cur_device_controller = DeviceController.get_for(cur_device)
            for _, cur_connections in cur_device_controller.absolute_connections.items():
                for cur_cnn in cur_connections:
                    if cur_cnn not in all_connections:
                        all_connections.append(cur_cnn)
        return all_connections

    def validate_inheritance(self):
        """
        This method validates that the inheritance of the related :class:`Setup`/:class:`Scenario` class was done
        correctly. It checks that all inner devices that are inherited has the same naming as their parents and also
        that every reused name (that is already be used for a device in the parent class) does also inherit from this
        parent scenario/setup device.

        In addition to that, it secures that either all devices are overwritten in the current class or no devices are
        overwritten in the related class.
        """

        # get parent scenario / setup class and check no multi inheritance
        parent_scenario_or_setup = self.get_next_parent_class()
        if parent_scenario_or_setup is None:
            # done, because the parent class is direct Scenario/Setup class
            return

        parent_scenario_or_setup_controller = self.__class__.get_for(parent_scenario_or_setup)

        devices = self.get_all_inner_device_classes()
        abs_parent_devices = parent_scenario_or_setup_controller.get_all_abs_inner_device_classes()
        abs_parent_devices_by_name = {cur_parent.__name__: cur_parent for cur_parent in abs_parent_devices}

        if len(devices) == 0:
            # ignore it because cur item has no own device definitions
            return

        # check that a device is newly defined or has the same name as the parent device
        for cur_item_device in devices:
            # check if name exists in parent
            relevant_parent_according_naming = abs_parent_devices_by_name.get(cur_item_device.__name__, None)

            # check if device is inherited from a parent
            relevant_parent_device_according_inheritance = None
            for cur_parent in abs_parent_devices:
                if issubclass(cur_item_device, cur_parent):
                    if relevant_parent_device_according_inheritance is not None:
                        # multi inheritance is not allowed
                        raise MultiInheritanceError(
                            f"found more than one {self._related_type.__name__}-Device parent classes for the "
                            f"class `{cur_item_device.__name__}` - multi inheritance is not allowed for device "
                            f"classes")
                    relevant_parent_device_according_inheritance = cur_parent

            # now check if both is fulfilled
            if relevant_parent_according_naming == relevant_parent_device_according_inheritance and \
                    relevant_parent_device_according_inheritance is not None:
                # device is inherited AND has the same name as used in parent -> ALLOWED
                pass
            elif relevant_parent_according_naming is None and relevant_parent_device_according_inheritance is None:
                # both are none -> it is a new device -> ALLOWED
                pass
            elif relevant_parent_according_naming is None:
                # reused a naming but does not inherit from it -> NOT ALLOWED
                raise DeviceOverwritingError(
                    f"the inner device class `{cur_item_device.__qualname__}` which inherits from another "
                    f"device `{relevant_parent_device_according_inheritance.__qualname__}` - it should also have "
                    f"the same name")
            elif relevant_parent_device_according_inheritance is None:
                # inherit from a parent device, but it doesn't have the same naming -> NOT ALLOWED
                raise DeviceOverwritingError(
                    f"the inner device class `{cur_item_device.__qualname__}` has the same name than the "
                    f"device `{relevant_parent_according_naming.__qualname__}` - it should also inherit from it")

        # secure that all parent devices are implemented here too
        for cur_parent in abs_parent_devices:
            found_parent = len([dev for dev in devices if issubclass(dev, cur_parent)]) > 0
            if not found_parent:
                raise DeviceOverwritingError(
                    f"found a device `{cur_parent.__qualname__}` which is part of a parent class, but it is "
                    f"not implemented in child class `{self.related_cls.__name__}`")

        # also check the parent class here
        self.__class__.get_for(parent_scenario_or_setup).validate_inheritance()

    def check_vdevice_feature_existence(self):
        """
        This method validates that the :class:`Feature` property set of a :class:`Device` holds all required
        :class:`Feature` objects of the related :class:`VDevice`. For this the method checks that every feature (that
        is used in a mapped :class:`VDevice`) also exists as a child :class:`Feature` property in the related
        :class:`Device` class.

        .. note::
            Variations are not related to this and will not be checked here.

        """

        for cur_device in self.get_all_abs_inner_device_classes():
            cur_device_instantiated_features = \
                DeviceController.get_for(cur_device).get_all_instantiated_feature_objects()
            for _, cur_feature in cur_device_instantiated_features.items():
                active_vdevice, related_device = cur_feature.active_vdevice_device_mapping
                if active_vdevice is None:
                    # ignore this, because no active vdevice exists
                    continue

                # secure that all the defined features in the VDevice also exist in the related device ->
                #  otherwise error
                orig_device_features = [
                    feat for _, feat in
                    DeviceController.get_for(related_device).get_all_instantiated_feature_objects().items()]
                active_vdevice_instantiated_features = \
                    VDeviceController.get_for(active_vdevice).get_all_instantiated_feature_objects()
                for _, cur_vdevice_feature in active_vdevice_instantiated_features.items():
                    # search for it
                    found_it = False
                    for cur_orig_feature in orig_device_features:
                        if isinstance(cur_orig_feature, cur_vdevice_feature.__class__):
                            found_it = True
                            break
                    if not found_it:
                        raise MissingFeaturesOfVDeviceError(
                            f"the device `{related_device.__name__}` which is mapped to the VDevice "
                            f"`{active_vdevice.__name__}` doesn't have an implementation for the feature "
                            f"`{cur_vdevice_feature.__class__.__name__}` required by the VDevice class "
                            f"`{active_vdevice.__name__}`")

    def get_absolute_single_connections(self) -> \
            Dict[Type[Device], Dict[str, Dict[Type[Device], Dict[str, List[Connection]]]]]:
        """
        This method determines the synchronized (both devices of a connection were updated) absolute connections
        between all devices of this scenario/setup.

        :return: returns a dictionary which provides all single connections between two devices that can be accessed
                 with `result[dev1][node-of-dev1][dev2][node-of-dev2]`
        """
        # start to generate the singles for every connection between the devices of every scenario
        all_abs_single_connections = {}

        all_devices = self.get_all_abs_inner_device_classes()
        for cur_from_device in all_devices:
            cur_from_device_controller = DeviceController.get_for(cur_from_device)

            # generate the whole `all_abs_single_connections` and convert the connections to singles
            for cur_to_device, cur_connections in cur_from_device_controller.absolute_connections.items():
                for cur_cnn in cur_connections:
                    if cur_from_device == cur_cnn.from_device:
                        cur_from_node = cur_cnn.from_node_name
                        cur_to_node = cur_cnn.to_node_name
                    else:
                        cur_from_node = cur_cnn.to_node_name
                        cur_to_node = cur_cnn.from_node_name
                    if cur_from_device not in all_abs_single_connections.keys():
                        all_abs_single_connections[cur_from_device] = {}
                    if cur_from_node not in all_abs_single_connections[cur_from_device].keys():
                        all_abs_single_connections[cur_from_device][cur_from_node] = {}
                    if cur_to_device not in \
                            all_abs_single_connections[cur_from_device][cur_from_node].keys():
                        all_abs_single_connections[cur_from_device][
                            cur_from_node][cur_to_device] = {}
                    if cur_to_node not in \
                            all_abs_single_connections[cur_from_device][cur_from_node][cur_to_device].keys():
                        all_abs_single_connections[cur_from_device][cur_from_node][cur_to_device][
                            cur_to_node] = cur_cnn.get_singles()
                        # we do not have to set the connection in communication device, because the absolute
                        # connections are always synchronized
                    else:
                        raise ValueError(
                            f'found multiple definitions for connection from device '
                            f'{cur_from_device.__qualname__} (node: `{cur_from_node}`) to device '
                            f'{cur_to_device.__qualname__} (node: `{cur_to_node}`) in scenario '
                            f'`{self.related_cls.__name__}`')
        return all_abs_single_connections

    def determine_raw_absolute_device_connections(self):
        """
        This method determines and creates the basic `_absolute_connections` for the related scenario/setup. Note,
        that this method only creates the class attribute and adds the synchronized connections (same on both sides if
        they are bidirectional). It does not analyse or take :class:`Feature` classes into consideration.
        """
        # determine next relevant base class
        next_base_class = self.get_next_parent_class()

        # executed this method for all parents too
        if next_base_class:
            NormalScenarioSetupController.get_for(next_base_class).determine_raw_absolute_device_connections()

        all_relevant_cnns = []

        all_devices = self.get_all_inner_device_classes()
        all_devices_as_strings = [search_device.__name__ for search_device in all_devices]

        # check if the devices of the current item has minimum one own connect() decorator
        has_connect_decorator = False
        for cur_device in all_devices:
            if len(DeviceController.get_for(cur_device).connections) > 0:
                has_connect_decorator = True

        if len(all_devices) == 0 and len(self.get_all_abs_inner_device_classes()) > 0:
            # only the parent class has defined scenarios -> use absolute data from next parent
            #  NOTHING TO DO, because we also use these devices in child setup/scenario
            return

        if len(all_devices) > 0 and not has_connect_decorator:
            # the current item has defined devices, but no own `@connect()` decorator -> use absolute data from
            #  next parent
            if next_base_class is not None:
                # only if there is a next base class
                next_base_class_controller = NormalScenarioSetupController.get_for(next_base_class)

                for cur_parent_cnn in next_base_class_controller.get_all_abs_connections():

                    # find all related devices (for this connection)
                    related_from_device = \
                        all_devices[all_devices_as_strings.index(cur_parent_cnn.from_device.__name__)]
                    related_to_device = all_devices[all_devices_as_strings.index(cur_parent_cnn.to_device.__name__)]
                    new_cnn = cur_parent_cnn.clone()
                    new_cnn.set_metadata_for_all_subitems(ConnectionMetadata(
                        from_device=related_from_device, from_device_node_name=cur_parent_cnn.from_node_name,
                        to_device=related_to_device, to_device_node_name=cur_parent_cnn.to_node_name)
                    )
                    all_relevant_cnns.append(new_cnn)

                # throw warning (but only if this scenario/setup has minimum one of the parent classes has inner
                # devices (and connection between them) by its own)
                if len(next_base_class_controller.get_all_abs_inner_device_classes()) > 0 and \
                        len(next_base_class_controller.get_all_abs_connections()) > 0:
                    logger.warning(
                        f"the collected `{self.related_cls.__name__}` class overwrites devices, but does "
                        f"not define connections between them by its own - please provide them in case you "
                        f"overwrite devices")
        else:
            # otherwise, use data from current layer, because there is no parent, no devices or this item overwrites
            # the connections from higher classes
            for cur_device in all_devices:
                for _, cur_cnn_list in DeviceController.get_for(cur_device).connections.items():
                    # now add every single connection correctly into the dictionary
                    all_relevant_cnns += [cur_cnn for cur_cnn in cur_cnn_list if cur_cnn not in all_relevant_cnns]

        # now set the absolute connections correctly
        for cur_cnn in all_relevant_cnns:
            DeviceController.get_for(cur_cnn.from_device).add_new_absolute_connection(cur_cnn)
            DeviceController.get_for(cur_cnn.to_device).add_new_absolute_connection(cur_cnn)



import balder
import balder.connections as cnns


class AddCalculateFeature(balder.Feature):

    all_numbers = []

    def get_numbers(self):
        """this method get all the single numbers that should be calculated"""
        raise NotImplementedError("has to be implemented in subclass")

    def add_numbers(self):
        """this method adds all the numbers"""
        raise NotImplementedError("has to be implemented in subclass")


class ProvidesANumberFeature(balder.Feature):
    the_number = 0

    def set_number(self, number):
        """user method that allows to set the number"""
        self.the_number = number

    def get_number(self):
        """returns the set number"""
        raise NotImplementedError("has to be implemented in subclass")

    def sends_the_number(self):
        """sends the set number"""
        raise NotImplementedError("has to be implemented in subclass")



from typing import Literal, Union, List
from datetime import datetime

import pathlib
import balder
import argparse
import logging
from multiprocessing import Queue
from .lib.utils import FixtureReturn

logger = logging.getLogger(__file__)


class MyTestException(Exception):
    pass


class RuntimeObserver:
    """This is a helper object, that will be used from this test environment to observe the execution order"""
    queue: Union[Queue, None] = None

    error_throwing = {}

    @staticmethod
    def add_entry(file: Union[str, pathlib.Path], cls: object, meth: callable, msg: str,
                  category: Literal["fixture", "testcase", "feature"] = None,
                  part: Literal["construction", "teardown"] = None):
        """
        adds a new entry and sends it over the queue

        :param file: the full filepath where the log will be generated

        :param cls: the class object, the entry is generated in

        :param meth: the method name, the entry is generated in

        :param msg: the message that should be inserted into the entry

        :param category: optional string of the category the entry is from

        :param part: optional string of the sub part the entry is from
        """
        if hasattr(meth, 'fn'):
            meth = meth.fn
        new_dataset = {
            "timestamp": datetime.now(), "file": file, "cls": "" if cls is None else cls.__name__,
            "meth": meth.__name__, "msg": msg, "category": category, "part": part
        }
        logger.info("{:22} | {:20} | {:30} | {:12} | {:15} | {}".format(
            pathlib.Path(file).parts[-1], "" if cls is None else cls.__name__, "" if meth is None else meth.__name__,
            "" if category is None else category, "" if part is None else part, "" if msg is None else msg))

        RuntimeObserver.queue.put(new_dataset)
        # check if we have to throw the error
        error_throwing_required = len(RuntimeObserver.error_throwing) > 0
        for cur_key in RuntimeObserver.error_throwing.keys():
            new_dataset_val = new_dataset[cur_key]
            if callable(new_dataset_val):
                new_dataset_val = new_dataset_val.__name__
            if new_dataset_val != RuntimeObserver.error_throwing[cur_key]:
                error_throwing_required = False
                break
        if error_throwing_required:
            raise MyTestException(f'raise test triggered exception for `{str(RuntimeObserver.error_throwing)}`')


class MyErrorThrowingPlugin(balder.BalderPlugin):
    """
    This is a plugin that reads the values from console arguments and sets these values into the
    :class:`RuntimeObserver`. The static method `RuntimeObserver.add_entry` will automatically throw an exception on the
    given position.
    """

    def addoption(self, argument_parser: argparse.ArgumentParser):
        argument_parser.add_argument('--test-error-file', help='the file id, the error should be thrown in')
        argument_parser.add_argument('--test-error-cls', help='the class id, the error should be thrown in')
        argument_parser.add_argument('--test-error-meth', help='the meth id, the error should be thrown in')
        argument_parser.add_argument('--test-error-part', help='the part (`construct` or `teardown`), the error should '
                                                               'be thrown in - only for fixtures')

    def modify_collected_pyfiles(self, pyfiles: List[pathlib.Path]) -> List[pathlib.Path]:
        # use this method to set the values
        RuntimeObserver.error_throwing = {}
        if self.balder_session.parsed_args.test_error_file:
            path = pathlib.Path(self.balder_session.parsed_args.test_error_file)
            if not path.is_absolute():
                path = str(self.balder_session.working_dir.joinpath(path))
            RuntimeObserver.error_throwing['file'] = path
        if self.balder_session.parsed_args.test_error_cls:
            RuntimeObserver.error_throwing['cls'] = self.balder_session.parsed_args.test_error_cls
        if self.balder_session.parsed_args.test_error_meth:
            RuntimeObserver.error_throwing['meth'] = self.balder_session.parsed_args.test_error_meth
        if self.balder_session.parsed_args.test_error_part:
            RuntimeObserver.error_throwing['part'] = self.balder_session.parsed_args.test_error_part
        return pyfiles

@balder.fixture(level="session")
def fixture_session():
    RuntimeObserver.add_entry(__file__, None, balderglob_fixture_session, "begin execution CONSTRUCTION of fixture",
                              category="fixture", part="construction")

    yield FixtureReturn.BALDERGLOB_SESSION

    RuntimeObserver.add_entry(__file__, None, balderglob_fixture_session, "begin execution TEARDOWN of fixture",
                              category="fixture", part="teardown")


@balder.fixture(level="setup")
def fixture_setup(fixture_session):
    RuntimeObserver.add_entry(__file__, None, balderglob_fixture_setup, "begin execution CONSTRUCTION of fixture",
                              category="fixture", part="construction")
    assert fixture_session == FixtureReturn.BALDERGLOB_SESSION

    yield FixtureReturn.BALDERGLOB_SETUP

    RuntimeObserver.add_entry(__file__, None, balderglob_fixture_setup, "begin execution TEARDOWN of fixture",
                              category="fixture", part="teardown")


@balder.fixture(level="scenario")
def fixture_scenario(fixture_session, fixture_setup):
    RuntimeObserver.add_entry(__file__, None, balderglob_fixture_scenario, "begin execution CONSTRUCTION of fixture",
                              category="fixture", part="construction")
    assert fixture_session == FixtureReturn.BALDERGLOB_SESSION
    assert fixture_setup == FixtureReturn.BALDERGLOB_SETUP

    yield FixtureReturn.BALDERGLOB_SCENARIO

    RuntimeObserver.add_entry(__file__, None, balderglob_fixture_scenario, "begin execution TEARDOWN of fixture",
                              category="fixture", part="teardown")


@balder.fixture(level="variation")
def fixture_variation(fixture_session, fixture_setup, fixture_scenario):
    RuntimeObserver.add_entry(__file__, None, balderglob_fixture_variation, "begin execution CONSTRUCTION of fixture",
                              category="fixture", part="construction")
    assert fixture_session == FixtureReturn.BALDERGLOB_SESSION
    assert fixture_setup == FixtureReturn.BALDERGLOB_SETUP
    assert fixture_scenario == FixtureReturn.BALDERGLOB_SCENARIO

    yield FixtureReturn.BALDERGLOB_VARIATION

    RuntimeObserver.add_entry(__file__, None, balderglob_fixture_variation, "begin execution TEARDOWN of fixture",
                              category="fixture", part="teardown")


@balder.fixture(level="testcase")
def fixture_testcase(fixture_session, fixture_setup, fixture_scenario, fixture_variation):
    RuntimeObserver.add_entry(__file__, None, balderglob_fixture_testcase, "begin execution CONSTRUCTION of fixture",
                              category="fixture", part="construction")
    assert fixture_session == FixtureReturn.BALDERGLOB_SESSION
    assert fixture_setup == FixtureReturn.BALDERGLOB_SETUP
    assert fixture_scenario == FixtureReturn.BALDERGLOB_SCENARIO
    assert fixture_variation == FixtureReturn.BALDERGLOB_VARIATION

    yield FixtureReturn.BALDERGLOB_TESTCASE

    RuntimeObserver.add_entry(__file__, None, balderglob_fixture_testcase, "begin execution TEARDOWN of fixture",
                              category="fixture", part="teardown")


# helper variable to give correct reference to add_entry
balderglob_fixture_session = fixture_session
balderglob_fixture_setup = fixture_setup
balderglob_fixture_scenario = fixture_scenario
balderglob_fixture_variation = fixture_variation
balderglob_fixture_testcase = fixture_testcase



import balder
import logging
from .setup_features import SetupFeatureIII, SetupFeatureIV
from ..lib.connections import BConnection
from ..balderglob import RuntimeObserver

logger = logging.getLogger(__name__)


class SetupB(balder.Setup):
    """This is a setup of category B (exactly the same as scenario B)"""

    class SetupDevice3(balder.Device):
        s_iii = SetupFeatureIII()

    @balder.connect(SetupDevice3, over_connection=BConnection)
    class SetupDevice4(balder.Device):
        s_iv = SetupFeatureIV()

    @balder.fixture(level="session")
    def fixture_session(self):
        RuntimeObserver.add_entry(__file__, SetupB, SetupB.fixture_session,  category="fixture", part="construction",
                                  msg="begin execution CONSTRUCTION of fixture")

        self.SetupDevice3.s_iii.do_something()
        self.SetupDevice4.s_iv.do_something()

        yield

        RuntimeObserver.add_entry(__file__, SetupB, SetupB.fixture_session,  category="fixture", part="teardown",
                                  msg="begin execution TEARDOWN of fixture")

        self.SetupDevice3.s_iii.do_something()
        self.SetupDevice4.s_iv.do_something()

    @balder.fixture(level="setup")
    def fixture_setup(self):
        RuntimeObserver.add_entry(__file__, SetupB, SetupB.fixture_setup,  category="fixture", part="construction",
                                  msg="begin execution CONSTRUCTION of fixture")

        self.SetupDevice3.s_iii.do_something()
        self.SetupDevice4.s_iv.do_something()

        yield

        RuntimeObserver.add_entry(__file__, SetupB, SetupB.fixture_setup,  category="fixture", part="teardown",
                                  msg="begin execution TEARDOWN of fixture")

        self.SetupDevice3.s_iii.do_something()
        self.SetupDevice4.s_iv.do_something()

    @balder.fixture(level="scenario")
    def fixture_scenario(self):
        RuntimeObserver.add_entry(__file__, SetupB, SetupB.fixture_scenario,  category="fixture", part="construction",
                                  msg="begin execution CONSTRUCTION of fixture")

        self.SetupDevice3.s_iii.do_something()
        self.SetupDevice4.s_iv.do_something()

        yield

        RuntimeObserver.add_entry(__file__, SetupB, SetupB.fixture_scenario,  category="fixture", part="teardown",
                                  msg="begin execution TEARDOWN of fixture")

        self.SetupDevice3.s_iii.do_something()
        self.SetupDevice4.s_iv.do_something()

    @balder.fixture(level="variation")
    def fixture_variation(self):
        RuntimeObserver.add_entry(__file__, SetupB, SetupB.fixture_variation,  category="fixture", part="construction",
                                  msg="begin execution CONSTRUCTION of fixture")

        self.SetupDevice3.s_iii.do_something()
        self.SetupDevice4.s_iv.do_something()

        yield

        RuntimeObserver.add_entry(__file__, SetupB, SetupB.fixture_variation,  category="fixture", part="teardown",
                                  msg="begin execution TEARDOWN of fixture")

        self.SetupDevice3.s_iii.do_something()
        self.SetupDevice4.s_iv.do_something()

    @balder.fixture(level="testcase")
    def fixture_testcase(self):
        RuntimeObserver.add_entry(__file__, SetupB, SetupB.fixture_testcase,  category="fixture", part="construction",
                                  msg="begin execution CONSTRUCTION of fixture")

        self.SetupDevice3.s_iii.do_something()
        self.SetupDevice4.s_iv.do_something()

        yield

        RuntimeObserver.add_entry(__file__, SetupB, SetupB.fixture_testcase,  category="fixture", part="teardown",
                                  msg="begin execution TEARDOWN of fixture")

        self.SetupDevice3.s_iii.do_something()
        self.SetupDevice4.s_iv.do_something()



import balder
from .features_setup import SendOverQueueFeature, ReceiveOverQueueFeature


class SetupQueueAliceToBob(balder.Setup):
    class Alice(balder.Device):
        send = SendOverQueueFeature()

    @balder.connect(with_device=Alice, over_connection=balder.Connection())
    class Bob(balder.Device):
        recv = ReceiveOverQueueFeature()



import balder
import logging

from ..lib.features import FeatureI, FeatureII
from ..lib.connections import AConnection
from ..balderglob import RuntimeObserver

logger = logging.getLogger(__file__)


class ScenarioAFactory:

    def get_scenario(self):

        class ScenarioA(balder.Scenario):
            """This is the scenario of category A"""

            class ScenarioDevice1(balder.Device):
                i = FeatureI()

            @balder.connect(ScenarioDevice1, over_connection=AConnection)
            class ScenarioDevice2(balder.Device):
                ii = FeatureII()

            def test_a_1(self):
                RuntimeObserver.add_entry(__file__, ScenarioA, ScenarioA.test_a_1, category="testcase",
                                          msg=f"execute Test `{ScenarioA.test_a_1.__qualname__}`")
                self.ScenarioDevice1.i.do_something()
                self.ScenarioDevice2.ii.do_something()

            def test_a_2(self):
                RuntimeObserver.add_entry(__file__, ScenarioA, ScenarioA.test_a_2, category="testcase",
                                          msg=f"execute Test `{ScenarioA.test_a_2.__qualname__}`")
                self.ScenarioDevice1.i.do_something()
                self.ScenarioDevice2.ii.do_something()

            @balder.fixture(level="session")
            def fixture_session(self):
                RuntimeObserver.add_entry(__file__, ScenarioA, ScenarioA.fixture_session, category="fixture",
                                          part="construction", msg="begin execution CONSTRUCTION of fixture")

                self.ScenarioDevice1.i.do_something()
                self.ScenarioDevice2.ii.do_something()

                yield

                RuntimeObserver.add_entry(__file__, ScenarioA, ScenarioA.fixture_session, category="fixture", part="teardown",
                                          msg="begin execution TEARDOWN of fixture")

                self.ScenarioDevice1.i.do_something()
                self.ScenarioDevice2.ii.do_something()

            @balder.fixture(level="setup")
            def fixture_setup(self):
                RuntimeObserver.add_entry(__file__, ScenarioA, ScenarioA.fixture_setup, category="fixture", part="construction",
                                          msg="begin execution CONSTRUCTION of fixture")
                self.ScenarioDevice1.i.do_something()
                self.ScenarioDevice2.ii.do_something()

                yield

                RuntimeObserver.add_entry(__file__, ScenarioA, ScenarioA.fixture_setup, category="fixture", part="teardown",
                                          msg="begin execution TEARDOWN of fixture")

                self.ScenarioDevice1.i.do_something()
                self.ScenarioDevice2.ii.do_something()

            @balder.fixture(level="scenario")
            def fixture_scenario(self):
                RuntimeObserver.add_entry(__file__, ScenarioA, ScenarioA.fixture_scenario, category="fixture",
                                          part="construction", msg="begin execution CONSTRUCTION of fixture")

                self.ScenarioDevice1.i.do_something()
                self.ScenarioDevice2.ii.do_something()

                yield

                RuntimeObserver.add_entry(__file__, ScenarioA, ScenarioA.fixture_scenario, category="fixture", part="teardown",
                                          msg="begin execution TEARDOWN of fixture")

                self.ScenarioDevice1.i.do_something()
                self.ScenarioDevice2.ii.do_something()

            @balder.fixture(level="variation")
            def fixture_variation(self):
                RuntimeObserver.add_entry(__file__, ScenarioA, ScenarioA.fixture_variation, category="fixture",
                                          part="construction", msg="begin execution CONSTRUCTION of fixture")
                self.ScenarioDevice1.i.do_something()
                self.ScenarioDevice2.ii.do_something()

                yield

                RuntimeObserver.add_entry(__file__, ScenarioA, ScenarioA.fixture_variation, category="fixture", part="teardown",
                                          msg="begin execution TEARDOWN of fixture")

                self.ScenarioDevice1.i.do_something()
                self.ScenarioDevice2.ii.do_something()

            @balder.fixture(level="testcase")
            def fixture_testcase(self):
                RuntimeObserver.add_entry(__file__, ScenarioA, ScenarioA.fixture_testcase, category="fixture",
                                          part="construction", msg="begin execution CONSTRUCTION of fixture")

                self.ScenarioDevice1.i.do_something()
                self.ScenarioDevice2.ii.do_something()

                yield

                RuntimeObserver.add_entry(__file__, ScenarioA, ScenarioA.fixture_testcase, category="fixture", part="teardown",
                                          msg="begin execution TEARDOWN of fixture")

                self.ScenarioDevice1.i.do_something()
                self.ScenarioDevice2.ii.do_something()

        return ScenarioA


factory = ScenarioAFactory()

ScenarioA1 = factory.get_scenario()



from __future__ import annotations
from typing import Union, Type, Callable, Generator, TYPE_CHECKING
import dataclasses

from .utils.typings import MethodLiteralType

if TYPE_CHECKING:
    from _balder.scenario import Scenario
    from _balder.setup import Setup


@dataclasses.dataclass
class FixtureMetadata:
    """
    describes meta information for a fixture
    """
    #: the namespace where it is defined (None if it is in a balderglob file)
    namespace: Union[None, Type[Scenario], Type[Setup]]
    #: the type of the fixture function
    function_type: MethodLiteralType
    #: the fixture callable itself
    callable: Callable
    #: the generator object (if the fixture is not a generator, it holds an empty generator)
    generator: Generator
    #: result according to the fixture's construction code (will be cleaned after it leaves a level)
    retval: object



from .feature_i import SetupFeatureI
from .feature_ii import SetupFeatureII
from .feature_i_overwritten import SetupFeatureIOverwritten
from .feature_new import SetupFeatureNew

__all__ = [
    "SetupFeatureI",
    "SetupFeatureII",
    "SetupFeatureIOverwritten",
    "SetupFeatureNew"
]



from __future__ import annotations
from typing import Type, Dict, List, Tuple, Union, Callable, Iterable, Any

import logging
import inspect
from collections import OrderedDict
from _balder.cnnrelations import OrConnectionRelation
from _balder.device import Device
from _balder.scenario import Scenario
from _balder.connection import Connection
from _balder.controllers.feature_controller import FeatureController
from _balder.controllers.device_controller import DeviceController
from _balder.controllers.normal_scenario_setup_controller import NormalScenarioSetupController
from _balder.parametrization import FeatureAccessSelector, Parameter
from _balder.exceptions import UnclearAssignableFeatureConnectionError, ConnectionIntersectionError, \
    MultiInheritanceError
from _balder.utils.functions import get_scenario_inheritance_list_of

logger = logging.getLogger(__file__)


class ScenarioController(NormalScenarioSetupController):
    """
    This is the controller class for :class:`Scenario` items.
    """

    # helper property to disable manual constructor creation
    __priv_instantiate_key = object()

    #: contains all existing scenarios and its corresponding controller object
    _items: Dict[Type[Scenario], ScenarioController] = {}

    _parametrization: Dict[Callable, Dict[str, Union[Iterable[Any], FeatureAccessSelector]]] = {}

    def __init__(self, related_cls, _priv_instantiate_key):

        # describes if the current controller is for setups or for scenarios (has to be set in child controller)
        self._related_type = Scenario

        # holds covered-by configuration
        self._covered_by = {}

        # this helps to make this constructor only possible inside the controller object
        if _priv_instantiate_key != ScenarioController.__priv_instantiate_key:
            raise RuntimeError('it is not allowed to instantiate a controller manually -> use the static method '
                               '`ScenarioController.get_for()` for it')

        if not isinstance(related_cls, type):
            raise TypeError('the attribute `related_cls` has to be a type (no object)')
        if not issubclass(related_cls, Scenario):
            raise TypeError(f'the attribute `related_cls` has to be a sub-type of `{Scenario.__name__}`')
        if related_cls == Scenario:
            raise TypeError(f'the attribute `related_cls` is `{Scenario.__name__}` - controllers for native type are '
                            f'forbidden')
        # contains a reference to the related class this controller instance belongs to
        self._related_cls = related_cls

    # ---------------------------------- STATIC METHODS ----------------------------------------------------------------

    @staticmethod
    def get_for(related_cls: Type[Scenario]) -> ScenarioController:
        """
        This class returns the current existing controller instance for the given item. If the instance does not exist
        yet, it will automatically create it and saves the instance in an internal dictionary.
        """
        if ScenarioController._items.get(related_cls) is None:
            item = ScenarioController(related_cls, _priv_instantiate_key=ScenarioController.__priv_instantiate_key)
            ScenarioController._items[related_cls] = item

        return ScenarioController._items.get(related_cls)

    # ---------------------------------- CLASS METHODS ----------------------------------------------------------------

    # ---------------------------------- PROPERTIES --------------------------------------------------------------------

    @property
    def related_cls(self) -> Type[Scenario]:
        return self._related_cls

    # ---------------------------------- PROTECTED METHODS -------------------------------------------------------------

    # ---------------------------------- METHODS -----------------------------------------------------------------------

    def register_parametrization(
            self,
            test_method: Callable,
            field_name: str,
            values: Iterable[Any] | FeatureAccessSelector
    ) -> None:
        """
        This method registers a custom parametrization for a test method of this Scenario
        """
        if test_method not in self.get_all_test_methods():
            raise ValueError(f'got test method `{test_method.__qualname__}` which is no part of the '
                             f'scenario `{self.related_cls}`')
        if test_method not in self._parametrization.keys():
            self._parametrization[test_method] = {}
        if field_name in self._parametrization[test_method].keys():
            raise ValueError(f'field name `{field_name}` for test method `{test_method.__qualname__}` already '
                             f'registered')
        self._parametrization[test_method][field_name] = values

    def get_parametrization_for(
            self,
            test_method: Callable,
            static: bool = True,
            dynamic: bool = True,
    ) -> OrderedDict[str, Iterable[Any] | FeatureAccessSelector] | None:
        """
        This method returns the parametrization for a test method of this Scenario. It returns the parameter
        configuration for every parameter in an OrderedDict.

        :param test_method: the test method of the Scenario
        :param static: if False, all static parameters will not be included into the dict.
        :param dynamic: if False, all dynamic parameters will not be included into the dict.
        """
        if test_method not in self._parametrization.keys():
            return None
        params = self._parametrization[test_method]

        # get arguments in defined order
        arguments = [name for name in inspect.getfullargspec(test_method).args if name in params.keys()]
        ordered_dict = OrderedDict()
        for cur_arg in arguments:
            cur_value = params[cur_arg]
            if isinstance(cur_value, FeatureAccessSelector) and dynamic is False:
                continue
            if not isinstance(cur_value, FeatureAccessSelector) and static is False:
                continue
            ordered_dict[cur_arg] = params[cur_arg]
        return ordered_dict

    def register_covered_by_for(self, meth: Union[str, None], covered_by: Union[Scenario, Callable, None]) -> None:
        """
        This method registers a covered-by statement for this Scenario. If `meth` is provided, the statement is for the
        specific test method of the scenario, otherwise it is for the whole setup. The item provided in `covered_by`
        describes the test object that covers this scenario (method).

        :param meth: if provided this attribute describes the test method that should be registered, otherwise the whole
                     scenario will be registered
        :param covered_by: describes the test object that covers this scenario (method)
        """
        if not (meth is None or isinstance(meth, str)):
            raise TypeError('meth needs to be None or a string')
        if meth is not None:
            if not meth.startswith('test_'):
                raise TypeError(
                    f"the use of the `@covered_by` decorator is only allowed for `Scenario` objects and test methods "
                    f"of `Scenario` objects - the method `{self.related_cls.__name__}.{meth}` does not start with "
                    f"`test_` and is not a valid test method")
            if not hasattr(self.related_cls, meth):
                raise ValueError(
                    f"the provided test method `{meth}` does not exist in scenario `{self.related_cls.__name__}`"
                )

        if meth not in self._covered_by.keys():
            self._covered_by[meth] = []
        if covered_by is None:
            # reset it
            # todo what if there are more than one decorator in one class
            del self._covered_by[meth]
        else:
            self._covered_by[meth].append(covered_by)

    def get_raw_covered_by_dict(self) -> Dict[Union[str, None], List[Union[Scenario, Callable]]]:
        """
        :return: returns the internal covered-by dictionary
        """
        return self._covered_by.copy()

    def get_abs_covered_by_dict(self) -> Dict[Union[str, None], List[Union[Scenario, Callable]]]:
        """
        This method resolves the covered-by statements over all inheritance levels. It automatically
        cleans up every inheritance of the covered_by decorators for every parent class of this scenario.
        """
        parent_classes = [p for p in self.related_cls.__bases__ if issubclass(p, Scenario) and p != Scenario]
        if len(parent_classes) > 1:
            raise MultiInheritanceError(
                f'can not resolve classes for `{self.related_cls}` because there are more than one Scenario based '
                f'parent classes'
            )
        # no more parent classes -> raw is absolute
        if len(parent_classes) == 0:
            return self.get_raw_covered_by_dict()
        parent_controller = self.__class__.get_for(parent_classes[0])
        self_raw_covered_by_dict = self.get_raw_covered_by_dict()

        #: first fill result with data from parent controller
        result = {
            k if k is None else getattr(self.related_cls, k.__name__): v
            for k, v in parent_controller.get_abs_covered_by_dict().items()
        }
        for cur_callable, cur_coveredby in self_raw_covered_by_dict.items():
            if cur_callable in result.keys():
                result[cur_callable].extend(cur_coveredby)
            else:
                result[cur_callable] = cur_coveredby
        return result

    def check_for_parameter_loop_in_dynamic_parametrization(self, cur_fn: Callable):
        """
        This method checks for a parameter loop in all dynamic parametrization for a specific test method. If it detects
        a loop an AttributeError is thrown
        """
        # only dynamic parametrization can have Parameter
        parametrization = self.get_parametrization_for(cur_fn, static=False, dynamic=True)

        def get_dependent_parameters_of_attribute(attribute: str) -> List[str] | None:
            cur_feature_access_selector = parametrization.get(attribute)
            if cur_feature_access_selector is None:
                return None
            # relevant are parameters only if they are from :class:`Parameter` and contained in the dynamic
            # parametrization
            return [param.name for param in cur_feature_access_selector.parameters.values()
                    if isinstance(param, Parameter) and param.name in parametrization.keys()]

        def recursive_parameter_loop_check(for_attribute, with_attribute: str):
            dependent_attr = get_dependent_parameters_of_attribute(with_attribute)
            if dependent_attr is None:
                # no problematic dependencies because attribute is no dynamic attribute
                return
            if len(dependent_attr) == 0:
                # no problematic dependencies
                return

            if for_attribute in dependent_attr:
                # loop detected
                raise AttributeError('detect a loop in Parameter() object - can not apply parametrization')
            # go deeper and resolve all dependent
            for cur_dependent_attr in dependent_attr:
                recursive_parameter_loop_check(for_attribute, cur_dependent_attr)
            return

        for cur_attr in parametrization.keys():
            recursive_parameter_loop_check(cur_attr, cur_attr)

    def get_next_parent_class(self) -> Union[Type[Scenario], None]:
        """
        This method returns the next parent class which is a subclass of the :class:`Scenario` itself.

        :return: returns the next parent class or None if the next parent class is :class:`Scenario`
                 itself
        """
        next_base_class = None
        for cur_base in self.related_cls.__bases__:
            if issubclass(cur_base, Scenario):
                if next_base_class is not None:
                    raise MultiInheritanceError(
                        f"found more than one Scenario parent classes for `{self.related_cls.__name__}` "
                        f"- multi inheritance is not allowed for Scenario/Setup classes")
                next_base_class = cur_base
        if next_base_class == Scenario:
            return None
        return next_base_class

    def get_all_test_methods(self) -> List[callable]:
        """
        This method returns all test methods that were defined in the related scenario. A testmethod has to start with
        `test_*`.
        """
        all_relevant_func = []

        all_methods = inspect.getmembers(self.related_cls, inspect.isfunction)
        for cur_method_name, cur_function in all_methods:
            if cur_method_name.startswith('test_'):
                all_relevant_func.append(cur_function)

        return all_relevant_func

    def get_ignore_test_methods(self) -> List[callable]:
        """
        This method returns a list of all methods that have the IGNORE marker. It automatically resolves marker that
        were provided on parent class scenarios.
        """
        result = []
        next_parent_class = get_scenario_inheritance_list_of(self.related_cls)[1]
        if next_parent_class != Scenario:
            next_parent_class_controller = ScenarioController.get_for(next_parent_class)
            next_parent_ignore_meths = next_parent_class_controller.get_ignore_test_methods()
            result.extend(next_parent_ignore_meths)
        for cur_ignore_meth_as_str in self.related_cls.IGNORE:
            cur_ignore_meth = getattr(self.related_cls, cur_ignore_meth_as_str)
            result.append(cur_ignore_meth)
        return list(set(result))

    def get_skip_test_methods(self) -> List[callable]:
        """
        This method returns a list of all methods that have the SKIP marker. It automatically resolves marker that were
        provided on parent class scenarios.
        """
        result = []
        next_parent_class = get_scenario_inheritance_list_of(self.related_cls)[1]
        next_parent_ignore_meths = []

        if next_parent_class != Scenario:
            next_parent_class_controller = ScenarioController.get_for(next_parent_class)
            next_parent_ignore_meths = next_parent_class_controller.get_ignore_test_methods()
            next_parent_skip_meths = next_parent_class_controller.get_skip_test_methods()
            result.extend(next_parent_skip_meths)

        for cur_skip_meth_as_str in self.related_cls.SKIP:
            cur_skip_meth = getattr(self.related_cls, cur_skip_meth_as_str)
            if cur_skip_meth in next_parent_ignore_meths:
                raise ValueError(f'found skip method `{cur_skip_meth}` defined in `{self.related_cls}.SKIP`, but was '
                                 f'already added to IGNORE in parent class')
            result.append(cur_skip_meth)

        return list(set(result))

    def get_run_test_methods(self) -> List[callable]:
        """
        This method returns a list of all methods that should run in this scenario. It automatically resolves
        SKIP/IGNORE marker that were provided on parent class scenarios.
        """
        result = (set(self.get_all_test_methods())
                  - set(self.get_skip_test_methods())
                  - set(self.get_ignore_test_methods()))
        return list(result)

    def validate_feature_clearance_for_parallel_connections(self):
        """
        This method validates for every active class-based feature (only the ones that have an active VDevice<->Device
        mapping), that there exist a clear scenario-device-connection for this feature. The method throws an
        :class:`UnclearAssignableFeatureConnectionError` if there exists more than one possible device-connection
        for the related devices and the method is not able to determine a clear connection.
        """

        for cur_from_device in self.get_all_abs_inner_device_classes():
            # determine all VDevice-Device mappings for this one, by iterating over all instantiated Feature classes
            cur_from_device_instantiated_features = \
                DeviceController.get_for(cur_from_device).get_all_instantiated_feature_objects()
            for _, cur_feature in cur_from_device_instantiated_features.items():
                mapped_vdevice, mapped_device = cur_feature.active_vdevice_device_mapping
                if mapped_device is None:
                    # ignore this, because we have no vDevices here
                    continue

                # now check if one or more single of the classbased connection are CONTAINED IN the possible
                # parallel connection (only if there exists more than one parallel)
                feature_cnn = FeatureController.get_for(
                    cur_feature.__class__).get_abs_class_based_for_vdevice()[mapped_vdevice]

                # search node names that is the relevant connection
                relevant_cnns: List[Connection] = []
                mapped_device_abs_cnns = DeviceController.get_for(mapped_device).get_all_absolute_connections()
                for _, all_connections in mapped_device_abs_cnns.items():
                    relevant_cnns += [cur_cnn for cur_cnn in all_connections
                                      if cur_cnn.has_connection_from_to(cur_from_device, end_device=mapped_device)]

                if len(relevant_cnns) <= 1:
                    # ignore if there are not more than one relevant connection
                    continue

                # there are some parallel connections -> check that only one fits with the feature
                matched_relevant_cnns = []
                for cur_relevant_cnn in relevant_cnns:
                    cur_relevant_cnn_singles = cur_relevant_cnn.get_singles()

                    for cur_relevant_single in cur_relevant_cnn_singles:
                        matches = [True for cur_feature_cnn in feature_cnn.get_singles()
                                   if cur_feature_cnn.contained_in(cur_relevant_single)]
                        if len(matches):
                            matched_relevant_cnns.append(True)
                            break
                if sum(matched_relevant_cnns) > 1:
                    raise UnclearAssignableFeatureConnectionError(
                        f"the devices {cur_from_device.__name__} and {mapped_device.__name__} have "
                        f"multiple parallel connections - the device `{cur_from_device.__name__}` uses a "
                        f"feature `{cur_feature.__class__.__name__}` that matches with the device "
                        f"`{mapped_device.__name__}`, but it is not clear which of the parallel connection "
                        f"could be used")

    def get_feature_cleaned_absolute_single_connections(self) -> \
            Tuple[Dict[Type[Device], Dict[str, Dict[Type[Device], Dict[str, List[Connection]]]]],  Dict[
                Tuple[Device, Device], List[Connection]]]:
        """
        This method returns all absolute-single connections between all devices of this scenario, but already cleaned
        based on the cumulated class-based decorators of all the feature devices.

        .. note::
            Please note, that the reduction candidates connections do not have to be unique.

        :return: returns a tuple with the cleaned up connections (sorted as dictionary per device) and the reduced
                 connections as second element
        """
        reduction_candidates = {}

        def add_reduction_candidate(device, other_device, connections: Connection):

            if (device, other_device) not in reduction_candidates.keys() and \
                    (other_device, device) not in reduction_candidates.keys():
                # we have to add it as new list
                reduction_candidates[(device, other_device)] = [connections]
            elif (device, other_device) in reduction_candidates.keys():
                reduction_candidates[(device, other_device)].append(connections)
            elif (other_device, device) in reduction_candidates.keys():
                reduction_candidates[(other_device, device)].append(connections)

        # start to generate the singles for every connection between the devices of every scenario
        all_abs_single_connections = self.get_absolute_single_connections()

        def reduce_based_on_feature_cnns_for_devices(
                feature_cnn: Connection, dev1: Type[Device], node_dev1: str, dev2: Type[Device], node_dev2: str):
            # execute further process only if there is exactly one relevant connection
            start_length_before_reduction = \
                len(all_abs_single_connections[dev1][node_dev1][
                        dev2][node_dev2])
            for cur_abs_connection in \
                    all_abs_single_connections[dev1][node_dev1][
                        dev2][node_dev2].copy():
                if not feature_cnn.contained_in(cur_abs_connection, ignore_metadata=True):
                    # this abs single connection is not fulfilled by the current feature -> remove it
                    all_abs_single_connections[dev1][node_dev1][
                        dev2][node_dev2].remove(cur_abs_connection)
                    add_reduction_candidate(dev1, dev2, cur_abs_connection)
            if start_length_before_reduction > 0 and \
                    len(all_abs_single_connections[dev1][node_dev1][
                            dev2][node_dev2]) == 0:
                raise ConnectionIntersectionError(
                    f"the `{self.related_cls.__name__}` has a connection from device "
                    f"`{dev1.__name__}` to `{dev2.__name__}` - some mapped VDevices of "
                    f"their feature classes define mismatched connections")

        def get_single_cnns_between_device_for_feature(from_device, to_device, relevant_feature_cnn):
            # search node names that is the relevant connection
            relevant_cnns: List[List[Connection]] = []
            for _, cur_node_data in all_abs_single_connections[from_device].items():
                for cur_to_device, cur_to_device_data in cur_node_data.items():
                    if cur_to_device == to_device:
                        relevant_cnns += [cur_cnns for _, cur_cnns in cur_to_device_data.items()]

            result_singles = None
            if len(relevant_cnns) > 1:
                # there exists parallel connections - filter only the relevant one
                for cur_single_cnns in relevant_cnns:
                    for cur_single_cnn in cur_single_cnns:
                        if relevant_feature_cnn.contained_in(cur_single_cnn):
                            # this is the relevant connection (all other can not fit, because we have
                            # already checked this with method
                            # `scenario_controller.validate_feature_clearance_for_parallel_connections()`)
                            result_singles = cur_single_cnns
                            break
                    if result_singles is not None:
                        break
            elif len(relevant_cnns) == 1:
                result_singles = relevant_cnns[0]
            if result_singles is None:
                raise ValueError("can not find relevant connection of all parallel connections")
            return result_singles

        all_devices = self.get_all_abs_inner_device_classes()
        for cur_from_device in all_devices:
            # determine all VDevice-Device mappings for this one, by iterating over all instantiated Feature classes
            cur_from_device_instantiated_features = \
                DeviceController.get_for(cur_from_device).get_all_instantiated_feature_objects()
            for _, cur_feature in cur_from_device_instantiated_features.items():
                mapped_vdevice, mapped_device = cur_feature.active_vdevice_device_mapping
                if mapped_device is None:
                    # ignore this, because we have no vDevices here
                    continue

                # now try to reduce the scenario connections according to the requirements of the feature class
                cur_feature_cnn = \
                    FeatureController.get_for(
                        cur_feature.__class__).get_abs_class_based_for_vdevice()[mapped_vdevice]

                device_cnn_singles = get_single_cnns_between_device_for_feature(
                    from_device=cur_from_device, to_device=mapped_device, relevant_feature_cnn=cur_feature_cnn)

                reduce_based_on_feature_cnns_for_devices(
                    cur_feature_cnn, device_cnn_singles[0].from_device, device_cnn_singles[0].from_node_name,
                    device_cnn_singles[0].to_device, device_cnn_singles[0].to_node_name)
                # do the same for the opposite direction (features are always bidirectional)
                reduce_based_on_feature_cnns_for_devices(
                    cur_feature_cnn, device_cnn_singles[0].to_device, device_cnn_singles[0].to_node_name,
                    device_cnn_singles[0].from_device, device_cnn_singles[0].from_node_name)

        return all_abs_single_connections, reduction_candidates

    def determine_absolute_device_connections(self):
        """
        This method determines the real possible Sub-Connections for every element of the scenarios. For this the method
        will create a possible intersection connection, for the :class:´Connection´ between two devices and
        all :class:`Connection`-Subtrees that are allowed for the mapped vDevices in the used :class:`Feature`
        classes.
        The data will be saved in the :class:`Device` property ``_absolute_connections``. If the method detects an empty
        intersection between two devices that are connected through a VDevice-Device mapping, the method will throw an
        exception.
        """

        # start to generate the singles for every connection between the devices of every scenario
        all_abs_single_connections, reduction_candidates = self.get_feature_cleaned_absolute_single_connections()

        # generate all required warnings
        for cur_warning_tuple in reduction_candidates:
            logger.warning(f"detect some connections between the devices `{cur_warning_tuple[0].__name__}` and "
                           f"`{cur_warning_tuple[1].__name__}` of scenario `{self.related_cls.__name__}` that can be "
                           f"reduced, because their related features only use a subset of the defined connection")

        # first cleanup the relevant absolute connections
        for cur_from_device, from_device_data in all_abs_single_connections.items():
            for _, from_node_data in from_device_data.items():
                for cur_to_device, to_device_data in from_node_data.items():
                    cur_from_device_controller = DeviceController.get_for(cur_from_device)
                    cur_to_device_controller = DeviceController.get_for(cur_to_device)

                    cur_from_device_controller.cleanup_absolute_connections_with(cur_to_device)
                    cur_to_device_controller.cleanup_absolute_connections_with(cur_from_device)

        # replace all absolute connection with the single ones
        for cur_from_device, from_device_data in all_abs_single_connections.items():
            for _, from_node_data in from_device_data.items():
                for cur_to_device, to_device_data in from_node_data.items():
                    for _, cur_single_cnns in to_device_data.items():

                        cur_from_device_controller = DeviceController.get_for(cur_from_device)
                        cur_to_device_controller = DeviceController.get_for(cur_to_device)

                        new_cnn = Connection.based_on(OrConnectionRelation(*cur_single_cnns))
                        new_cnn.set_metadata_for_all_subitems(cur_single_cnns[0].metadata)
                        if cur_from_device == cur_single_cnns[0].from_device:
                            cur_from_device_controller.add_new_absolute_connection(new_cnn)
                        else:
                            cur_to_device_controller.add_new_absolute_connection(new_cnn)
